import pytest

from cyclonedds.core import Listener, Qos, Policy
from cyclonedds.util import duration, timestamp
from cyclonedds.domain import DomainParticipant
from cyclonedds.pub import Publisher, DataWriter
from cyclonedds.sub import Subscriber, DataReader
from cyclonedds.topic import Topic
from support_modules.testtopics import Message
from time import sleep


def test_on_data_available(manual_setup, hitpoint):
    class L(Listener):
        def on_data_available(self, _):
            hitpoint.hit()

    manual_setup.dr(listener=L())
    manual_setup.dw().write(manual_setup.msg)

    assert hitpoint.was_hit()


def test_data_available_listeners(manual_setup, hitpoint_factory):
    hpf = hitpoint_factory

    class MyListener(Listener):
        def __init__(self):
            super().__init__()
            self.hitpoint_data_available = hpf()
            self.hitpoint_pub_matched = hpf()
            self.hitpoint_sub_matched = hpf()

        def on_data_available(self, reader):
            self.hitpoint_data_available.hit()

        def on_publication_matched(self, writer,status):
            self.hitpoint_pub_matched.hit()

        def on_subscription_matched(self, reader, status):
            self.hitpoint_sub_matched.hit()

    domain_participant_listener = MyListener()
    manual_setup.dp.listener = domain_participant_listener

    publisher_listener = MyListener()
    manual_setup.pub(listener=publisher_listener)

    subscriber_listener = MyListener()
    manual_setup.sub(listener=subscriber_listener)

    datawriter_listener = MyListener()
    datawriter = manual_setup.dw(listener=datawriter_listener)

    datareader_listener = MyListener()
    manual_setup.dr(listener=datareader_listener)

    datawriter.write(manual_setup.msg)

    #  Assertions, _only_ datawriter should publication match,
    # _only_ datareader should subscriber match and receive data

    assert datawriter_listener.hitpoint_pub_matched.was_hit()
    assert datareader_listener.hitpoint_sub_matched.was_hit()
    assert datareader_listener.hitpoint_data_available.was_hit()

    assert domain_participant_listener.hitpoint_pub_matched.was_not_hit()
    assert domain_participant_listener.hitpoint_sub_matched.was_not_hit()
    assert domain_participant_listener.hitpoint_data_available.was_not_hit()

    assert publisher_listener.hitpoint_pub_matched.was_not_hit()
    assert publisher_listener.hitpoint_sub_matched.was_not_hit()
    assert publisher_listener.hitpoint_data_available.was_not_hit()

    assert subscriber_listener.hitpoint_pub_matched.was_not_hit()
    assert subscriber_listener.hitpoint_sub_matched.was_not_hit()
    assert subscriber_listener.hitpoint_data_available.was_not_hit()

    assert datawriter_listener.hitpoint_sub_matched.was_not_hit()
    assert datawriter_listener.hitpoint_data_available.was_not_hit()

    assert datareader_listener.hitpoint_pub_matched.was_not_hit()


def test_on_offered_deadline_missed(manual_setup, hitpoint):
    class MyListener(Listener):
        def on_offered_deadline_missed(self, writer, status):
            hitpoint.hit(data=timestamp.now())

    wqos = Qos(Policy.Deadline(duration(seconds=0.2)))
    datawriter = manual_setup.dw(qos=wqos, listener=MyListener())
    manual_setup.dr()

    write_time = timestamp.now()
    datawriter.write(manual_setup.msg)
    sleep(1)

    assert hitpoint.was_hit()
    delay = hitpoint.data - write_time
    assert delay >= duration(seconds=0.2)  


def test_on_offered_incompatible_qos(manual_setup, hitpoint):
    class MyListener(Listener):
        def on_offered_incompatible_qos(self, writer, status):
            hitpoint.hit()

    wqos = Qos(Policy.Durability.Volatile)
    rqos = Qos(Policy.Durability.Transient)

    datawriter = manual_setup.dw(qos=wqos, listener=MyListener())
    manual_setup.dr(qos=rqos)

    datawriter.write(manual_setup.msg)

    assert hitpoint.was_hit()


def test_on_requested_incompatible_qos(manual_setup, hitpoint):
    class MyListener(Listener):
        def on_requested_incompatible_qos(self, writer, status):
            hitpoint.hit()

    wqos = Qos(Policy.Durability.Volatile)
    rqos = Qos(Policy.Durability.Transient)

    manual_setup.dr(qos=rqos, listener=MyListener())
    datawriter = manual_setup.dw(qos=wqos)

    datawriter.write(manual_setup.msg)

    assert hitpoint.was_hit()


def test_liveliness(manual_setup, hitpoint_factory):
    handler = hitpoint_factory()
    alive = hitpoint_factory()
    notalive = hitpoint_factory()

    class MyListenerWriter(Listener):
        def on_liveliness_lost(self, writer, status):
            handler.hit(data=timestamp.now())

    class MyListenerReader(Listener):
        def on_liveliness_changed(self, reader, status):
            if status.alive_count == 1:
                alive.hit(data=status.alive_count_change)
            else:
                notalive.hit(data=status.alive_count_change)

    qos = Qos(
        Policy.Liveliness.ManualByTopic(duration(seconds=0.2)),
        Policy.Ownership.Exclusive
    )

    manual_setup.tp(qos=qos)
    manual_setup.dr(listener=MyListenerReader())
    datawriter = manual_setup.dw(listener=MyListenerWriter())

    write_time = timestamp.now()
    datawriter.write(manual_setup.msg)

    assert handler.was_hit()
    assert handler.data - write_time >= duration(seconds=0.2)
    assert alive.was_hit() and alive.data == 1
    assert notalive.was_hit() and notalive.data == -1


def test_on_requested_deadline_missed(manual_setup, hitpoint):
    class MyListener(Listener):
        def on_requested_deadline_missed(self, reader, status):
            hitpoint.hit(data=timestamp.now())

    qos = Qos(Policy.Deadline(duration(seconds=0.2)))
    datawriter = manual_setup.dw(qos=qos)
    manual_setup.dr(qos=qos, listener=MyListener())

    write_time = timestamp.now()
    datawriter.write(manual_setup.msg)
    sleep(1)

    assert hitpoint.was_hit()
    assert hitpoint.data - write_time >= duration(seconds=0.2)


def test_on_sample_rejected(manual_setup, hitpoint):
    class MyListener(Listener):
        def on_sample_rejected(self, reader, status):
            hitpoint.hit()

    qos = Qos(
        Policy.ResourceLimits(max_samples=1, max_instances=1, max_samples_per_instance=1),
        Policy.Durability.Transient,
        Policy.History.KeepAll
    )

    datawriter = manual_setup.dw(qos=qos, listener=MyListener())
    manual_setup.dr(qos=qos, listener=MyListener())

    datawriter.write(manual_setup.msg)
    assert hitpoint.was_not_hit()

    datawriter.write(manual_setup.msg2)
    assert hitpoint.was_hit()


def test_on_sample_lost(manual_setup, hitpoint):
    class MyListener(Listener):
        def on_sample_lost(self, reader, status):
            hitpoint.hit()

    qos = Qos(Policy.DestinationOrder.BySourceTimestamp)

    datawriter = manual_setup.dw(qos=qos)
    datareader = manual_setup.dr(qos=qos, listener=MyListener())

    t1 = timestamp.now()
    t2 = t1 + duration(seconds=1)

    datawriter.write(manual_setup.msg, timestamp=t2)
    datareader.take()
    datawriter.write(manual_setup.msg, timestamp=t1)

    assert hitpoint.was_hit()



def test_pub_matched_cross_participant(hitpoint_factory):
    hpf = hitpoint_factory

    class MyListener(Listener):
        def __init__(self):
            super().__init__()
            self.hitpoint_data_available = hpf()
            self.hitpoint_pub_matched = hpf()
            self.hitpoint_sub_matched = hpf()

        def on_data_available(self, reader):
            self.hitpoint_data_available.hit()

        def on_publication_matched(self, writer,status):
            self.hitpoint_pub_matched.hit()

        def on_subscription_matched(self, reader, status):
            self.hitpoint_sub_matched.hit()

    domain1_participant_listener = MyListener()
    domain2_participant_listener = MyListener()
    dp1 = DomainParticipant(listener=domain1_participant_listener)
    dp2 = DomainParticipant(listener=domain2_participant_listener)

    tp1 = Topic(dp1, "Message", Message)
    tp2 = Topic(dp2, "Message", Message)

    publisher_listener = MyListener()
    pub = Publisher(dp1, listener=publisher_listener)

    subscriber_listener = MyListener()
    sub = Subscriber(dp2, listener=subscriber_listener)

    datawriter_listener = MyListener()
    datawriter = DataWriter(pub, tp1, listener=datawriter_listener)

    datareader_listener = MyListener()
    datareader = DataReader(sub, tp2, listener=datareader_listener)

    datawriter.write(Message("hi!"))

    #  Assertions, _only_ datawriter should publication match,
    # _only_ datareader should subscriber match and receive data

    assert datawriter_listener.hitpoint_pub_matched.was_hit()
    assert datareader_listener.hitpoint_sub_matched.was_hit()
    assert datareader_listener.hitpoint_data_available.was_hit()

    assert domain1_participant_listener.hitpoint_pub_matched.was_not_hit()
    assert domain1_participant_listener.hitpoint_sub_matched.was_not_hit()
    assert domain1_participant_listener.hitpoint_data_available.was_not_hit()
    assert domain2_participant_listener.hitpoint_pub_matched.was_not_hit()
    assert domain2_participant_listener.hitpoint_sub_matched.was_not_hit()
    assert domain2_participant_listener.hitpoint_data_available.was_not_hit()

    assert publisher_listener.hitpoint_pub_matched.was_not_hit()
    assert publisher_listener.hitpoint_sub_matched.was_not_hit()
    assert publisher_listener.hitpoint_data_available.was_not_hit()

    assert subscriber_listener.hitpoint_pub_matched.was_not_hit()
    assert subscriber_listener.hitpoint_sub_matched.was_not_hit()
    assert subscriber_listener.hitpoint_data_available.was_not_hit()

    assert datawriter_listener.hitpoint_sub_matched.was_not_hit()
    assert datawriter_listener.hitpoint_data_available.was_not_hit()

    assert datareader_listener.hitpoint_pub_matched.was_not_hit()

def test_listener_assignment_exhaustively():
    evt_names = [
        "on_data_available",
        "on_inconsistent_topic",
        "on_liveliness_lost",
        "on_liveliness_changed",
        "on_offered_deadline_missed",
        "on_offered_incompatible_qos",
        "on_data_on_readers",
        "on_sample_lost",
        "on_sample_rejected",
        "on_requested_deadline_missed",
        "on_requested_incompatible_qos",
        "on_publication_matched",
        "on_subscription_matched",
    ]

    funcs = {}
    for n in evt_names:
        def handler(*args, **kwargs):
            print(*args, **kwargs)
        funcs[n] = handler

    # Check keyword arguments assign to the Listener instance
    l = Listener(**funcs)
    for n in evt_names:
        assert getattr(l, n) is funcs[n]

    # Check the set_on functions assign to the listener instance
    funcs2 = {}
    for n in evt_names:
        def handler2(*args, **kwargs):
            print(*args, **kwargs)
        funcs2[n] = handler2
        getattr(l, f"set_{n}")(handler2)

    for n in evt_names:
        assert getattr(l, n) is not funcs[n]
        assert getattr(l, n) is funcs2[n]



from inspect import isclass, getmro
from textwrap import indent
from typing import Any, Type, List

from cyclonedds.idl import IdlStruct, IdlUnion, IdlBitmask, IdlEnum, types as pt
from cyclonedds.idl._type_normalize import (
    get_idl_annotations,
    get_idl_field_annotations,
    get_extended_type_hints,
    WrapOpt,
)
from cyclonedds.idl._type_helper import get_origin, get_args
from cyclonedds.idl.types import (
    array,
    bounded_str,
    sequence,
    _type_code_align_size_default_mapping,
    NoneType,
    char,
    typedef,
    case,
    default,
)


class _State:
    def __init__(self) -> None:
        self.types = set()
        self.output = ""
        self.module_scope = []

    def set_mod_scope(self, scope: List[str]):
        while len(scope) < len(self.module_scope):
            self._exit_scope()

        if not self.module_scope:
            for i in range(len(scope)):
                self._enter_scope(scope[i])
            return

        for i in range(len(self.module_scope)):
            if scope[i] != self.module_scope[i]:
                for j in range(i, len(self.module_scope)):
                    self._exit_scope()
                for j in range(i, len(scope)):
                    self._enter_scope(scope[j])
                break

    def _exit_scope(self):
        indent = "    " * (len(self.module_scope) - 1)
        self.output += f"{indent}}};\n"
        self.module_scope.pop()

    def _enter_scope(self, scope):
        indent = "    " * len(self.module_scope)
        self.output += f"{indent}module {scope} {{\n"
        self.module_scope.append(scope)

    def add_output(self, data):
        self.output += indent(data, "    " * len(self.module_scope))


class IdlType:
    easy_types = {
        pt.int8: "char",
        pt.int16: "short",
        pt.int32: "long",
        pt.int64: "long long",
        pt.uint8: "octet",
        pt.uint16: "short",
        pt.uint32: "unsigned long",
        pt.uint64: "unsigned long long",
        pt.float32: "float",
        pt.float64: "double",
        pt.wchar: "wchar",
        int: "long long",
        bool: "bool",
        float: "double",
        char: "char",
        str: "string",
    }

    @classmethod
    def _kind_type(cls, state, _type):
        if _type in cls.easy_types:
            return cls.easy_types[_type]
        elif _type in _type_code_align_size_default_mapping:
            return get_args(_type)[1]
        elif isinstance(_type, WrapOpt):
            return "@optional " + cls._kind_type(state, _type.inner)
        elif isclass(_type) and issubclass(
            _type, (IdlStruct, IdlUnion, IdlBitmask, IdlEnum)
        ):
            cls._proc_type(state, _type)
            return _type.__idl_typename__.replace(".", "::")
        elif get_origin(_type) == list:
            return "sequence<" + cls._kind_type(state, get_args(_type)[0]) + "> "
        elif get_origin(_type) == dict:
            return (
                "map<"
                + cls._kind_type(state, get_args(_type)[0])
                + ", "
                + cls._kind_type(state, get_args(_type)[0])
                + "> "
            )
        elif isinstance(_type, typedef):
            cls._proc_type(state, _type)
            return _type.name
        elif isinstance(_type, array):
            return cls._kind_type(state, _type.subtype)
        elif isinstance(_type, sequence):
            if _type.max_length:
                return (
                    "sequence<"
                    + cls._kind_type(state, _type.subtype)
                    + f", {_type.max_length}> "
                )
            else:
                return "sequence<" + cls._kind_type(state, _type.subtype) + "> "
        elif isinstance(_type, bounded_str):
            return f"string<{_type.max_length}> "

        return "unknown"

    @classmethod
    def _array_size(cls, _type):
        if isinstance(_type, array):
            inner = cls._array_size(_type.subtype)
            if inner is not None:
                return (_type.length, *inner)
            return (_type.length,)
        return tuple()

    @classmethod
    def _scoped_name(cls, name):
        m = name.replace(".", "::").split("::")
        return m[:-1], m[-1]

    @classmethod
    def _annot(cls, _type):
        out = ""
        for name, value in get_idl_annotations(_type).items():
            if name == "nested":
                out += "@nested\n"
            elif name == "autoid":
                out += f"@autoid({value.upper()})\n"
            elif name == "extensibility":
                out += f"@{value}\n"
            elif name == "bit_bound":
                out += f"@bit_bound({value})"
            else:
                out += f"/* @{name}({value}) */"
        return out

    @classmethod
    def _proc_type(cls, state, _type, top_level=False):
        if id(_type) in state.types:
            return

        state.types.add(id(_type))

        if isclass(_type) and issubclass(_type, IdlStruct):
            out = cls._annot(_type)

            scope, enname = cls._scoped_name(_type.__idl__.idl_transformed_typename)
            out += f"struct {enname} "
            basefields = []
            for base in getmro(_type)[1:]:
                if not issubclass(base, IdlStruct) or base is IdlStruct:
                    continue

                cls._proc_type(state, base)
                basefields.extend(get_extended_type_hints(base).keys())
                _, basename = cls._scoped_name(base.__idl__.idl_transformed_typename)
                out += f": {basename} "
                break
            out += "{"
            field_annot = get_idl_field_annotations(_type)
            for name, _type in get_extended_type_hints(_type).items():
                if name in basefields:
                    continue
                out += "\n    "
                if "key" in field_annot.get(name, {}):
                    out += "@key "
                realName = name
                if name in field_annot and "name" in field_annot[name]:
                    realName = field_annot[name]["name"]
                out += cls._kind_type(state, _type) + f" {realName}"
                for arr in cls._array_size(_type):
                    out += f"[{arr}]"
                out += ";"
            out += "\n};\n"
            state.set_mod_scope(scope)
            state.add_output(out)
        elif isclass(_type) and issubclass(_type, IdlUnion):
            out = cls._annot(_type)

            scope, enname = cls._scoped_name(_type.__idl__.idl_transformed_typename)
            out += f"union {enname}"
            out += (
                " switch (" + cls._kind_type(state, _type.__idl_discriminator__) + ") {"
            )
            field_annot = get_idl_field_annotations(_type)
            for name, __type in get_extended_type_hints(_type).items():
                if isinstance(__type, pt.case):
                    for l in __type.labels:
                        if isclass(_type.__idl_discriminator__) and issubclass(
                            _type.__idl_discriminator__, IdlEnum
                        ):
                            dscope, _ = cls._scoped_name(
                                _type.__idl_discriminator__.__idl_typename__
                            )
                            enum_scope = "::".join(dscope)
                            out += f"\n    case {enum_scope}::{_type.__idl_discriminator__(l).name}:"
                        else:
                            out += f"\n    case {l}:"
                else:
                    out += "\n    default:"

                realName = name
                if name in field_annot and "name" in field_annot[name]:
                    realName = field_annot[name]["name"]
                out += "\n        " + cls._kind_type(state, __type.subtype) + f" {realName}"
                for arr in cls._array_size(__type.subtype):
                    out += f"[{arr}]"
                out += ";"
            out += "\n};\n"
            state.set_mod_scope(scope)
            state.add_output(out)
        elif isclass(_type) and issubclass(_type, IdlBitmask):
            out = cls._annot(_type)

            scope, enname = cls._scoped_name(_type.__idl_typename__)
            out += f"bitmask {enname} {{"
            for name, __type in get_extended_type_hints(_type).items():
                out += f"\n    @position({_type.__idl_positions__[name]}) {name},"
            out = out[:-1]
            out += f"\n}};\n"
            state.set_mod_scope(scope)
            state.add_output(out)
        elif isclass(_type) and issubclass(_type, IdlEnum):
            out = cls._annot(_type)

            scope, enname = cls._scoped_name(_type.__idl_typename__)
            out += f"enum {enname} {{"
            v = -1
            for f in _type:
                if f.value != v + 1:
                    out += f"\n    @value({f.value}) {f.name},"
                else:
                    out += f"\n    {f.name},"
                v = f.value
            out = out[:-1]
            out += f"\n}};\n"
            state.set_mod_scope(scope)
            state.add_output(out)
        elif isinstance(_type, typedef):
            scope, enname = cls._scoped_name(_type.name)
            out = "typedef " + cls._kind_type(state, _type.subtype) + f" {_type.name}"
            for arr in cls._array_size(_type):
                out += f"[{arr}]"
            out += ";\n"
            state.set_mod_scope(scope)
            state.add_output(out)

    @classmethod
    def idl(cls, types: Type[Any]) -> str:
        state = _State()
        for _type in types:
            cls._proc_type(state, _type)
        state.set_mod_scope([])
        return state.output.strip()



from dataclasses import dataclass
from cyclonedds.qos import Policy
from cyclonedds.internal import dds_infinity
from rich.pretty import Pretty
from rich.text import Text
from rich.console import group
from datetime import timedelta


def fmtime(nanoseconds: int) -> str:
    if nanoseconds == dds_infinity:
        return "infinity"
    if nanoseconds == 0:
        return "zero"

    microseconds = nanoseconds // 1000
    nanoseconds -= microseconds * 1000
    milliseconds = microseconds // 1000
    microseconds -= milliseconds * 1000
    seconds = milliseconds // 1000
    milliseconds -= seconds * 1000
    minutes = seconds // 60
    seconds -= minutes * 60
    hours = minutes // 60
    minutes -= hours * 60
    days = hours // 24
    hours -= days * 24

    ret = "" if days == 0 else f"{days} day{'s' if days > 0 else ''}, "
    ret += "" if hours == 0 else f"{hours} hour{'s' if hours > 0 else ''}, "
    ret += "" if minutes == 0 else f"{minutes} minute{'s' if minutes > 0 else ''}, "
    ret += "" if seconds == 0 else f"{seconds} second{'s' if seconds > 0 else ''}, "
    ret += (
        ""
        if milliseconds == 0
        else f"{milliseconds} millisecond{'s' if milliseconds > 0 else ''}, "
    )
    ret += (
        ""
        if microseconds == 0
        else f"{microseconds} microsecond{'s' if microseconds > 0 else ''}, "
    )
    ret += (
        ""
        if nanoseconds == 0
        else f"{nanoseconds} nanosecond{'s' if nanoseconds > 0 else ''}, "
    )

    ret = ret[:-2]

    return ret


@dataclass
class Property:
    key: str
    value: str


@dataclass
class BinaryProperty:
    key: str
    value: bytes


def rich_format_policy(policy: Policy):
    if policy.__scope__ in [
        "Durability",
        "Ownership",
        "DestinationOrder",
        "IgnoreLocal",
    ] or policy in [Policy.Reliability.BestEffort, Policy.History.KeepAll]:
        return f"[bold magenta]{policy.__class__.__name__}[/]"
    if isinstance(policy, Policy.Reliability.Reliable):
        return Pretty(
            Policy.Reliability.Reliable(
                max_blocking_time=fmtime(policy.max_blocking_time)
            )
        )
    if isinstance(policy, Policy.Lifespan):
        return Pretty(Policy.Lifespan(lifespan=fmtime(policy.lifespan)))
    if isinstance(policy, Policy.Deadline):
        return Pretty(Policy.Deadline(deadline=fmtime(policy.deadline)))
    if isinstance(policy, Policy.LatencyBudget):
        return Pretty(Policy.LatencyBudget(budget=fmtime(policy.budget)))
    if policy.__scope__ == "Liveliness":
        return Pretty(policy.__class__(lease_duration=fmtime(policy.lease_duration)))
    if isinstance(policy, Policy.TimeBasedFilter):
        return Pretty(Policy.TimeBasedFilter(filter_time=fmtime(policy.filter_time)))
    if isinstance(policy, Policy.ReaderDataLifecycle):
        return Pretty(
            Policy.ReaderDataLifecycle(
                autopurge_nowriter_samples_delay=fmtime(
                    policy.autopurge_nowriter_samples_delay
                ),
                autopurge_disposed_samples_delay=fmtime(
                    policy.autopurge_disposed_samples_delay
                ),
            )
        )
    if isinstance(policy, Policy.Property):
        return Pretty(Property(key=policy.key, value=policy.value))
    if isinstance(policy, Policy.BinaryProperty):
        return Pretty(BinaryProperty(key=policy.key, value=policy.value))
    return Pretty(policy)


@group()
def rich_qos(qos):
    for policy in qos:
        yield rich_format_policy(policy)



from cyclonedds.domain import DomainParticipant
from cyclonedds.topic import Topic
from cyclonedds.pub import DataWriter
from cyclonedds.idl import IdlEnum, IdlUnion
from cyclonedds.idl.types import case


class MyEnum(IdlEnum):
    A = 1
    B = 0


class MyUnion(IdlUnion, discriminator=MyEnum):
    a: case[MyEnum.A, int]
    b: case[MyEnum.B, str]


def test_enum_as_discriminator():
    dp = DomainParticipant(0)
    tp = Topic(dp, 'EnumDiscrTopic', MyUnion)
    dw = DataWriter(dp, tp)
    dw.write(MyUnion(a=9))
    dw.write(MyUnion(b="Hello"))



import pytest

from cyclonedds.domain import DomainParticipant
from cyclonedds.topic import Topic
from cyclonedds.sub import DataReader
from cyclonedds.pub import DataWriter

from support_modules.testtopics import KeyedArrayType


def test_keyed_type_alignment():
    dp = DomainParticipant()
    tp = Topic(dp, "Test", KeyedArrayType)
    dw = DataWriter(dp, tp)
    dr = DataReader(dp, tp)

    samp1 = KeyedArrayType(
        [-1] * 3,
        [-1] * 3,
        [-1] * 3,
        [-1] * 3,
        [-1] * 3,
        [-1] * 3
    )

    dw.write(samp1)
    samp2 = dr.read()[0]
    assert KeyedArrayType.__idl__.serialize_key_normalized(samp1) == KeyedArrayType.__idl__.serialize_key_normalized(samp2)



"""
  Generated by Eclipse Cyclone DDS idlc Python Backend
  Cyclone DDS IDL version: v0.11.0
  Module: DDS.XTypes
  IDL file: ddsi_xt_typemap.idl

"""

from dataclasses import dataclass
from enum import auto
from typing import TYPE_CHECKING, Optional

import cyclonedds.idl as idl
import cyclonedds.idl.annotations as annotate
import cyclonedds.idl.types as types

# root module import for resolving types
import cyclonedds.idl._typesupport.DDS


@dataclass
@annotate.final
@annotate.autoid("sequential")
class TypeMapping(idl.IdlStruct, typename="DDS.XTypes.TypeMapping"):
    identifier_object_pair_minimal: types.sequence['cyclonedds.idl._typesupport.DDS.XTypes.TypeIdentifierTypeObjectPair']
    identifier_object_pair_complete: types.sequence['cyclonedds.idl._typesupport.DDS.XTypes.TypeIdentifierTypeObjectPair']
    identifier_complete_minimal: types.sequence['cyclonedds.idl._typesupport.DDS.XTypes.TypeIdentifierPair']





import pytest

from cyclonedds.core import Entity, Listener, DDSException
from cyclonedds.domain import DomainParticipant
from cyclonedds.topic import Topic
from cyclonedds.sub import Subscriber, DataReader
from cyclonedds.pub import Publisher, DataWriter
from cyclonedds.util import isgoodentity

from support_modules.testtopics import Message


def test_create_entity():
    dp = DomainParticipant(0)

    assert isgoodentity(dp)


def test_delete_entity():
    dp = DomainParticipant(0)

    assert dp.get_participant() == dp

    reference = dp._ref
    del dp

    with pytest.raises(DDSException):
        Entity(reference).get_participant()


def test_get_parent():
    dp = DomainParticipant(0)

    assert dp.get_parent() == dp.parent == None

    tp = Topic(dp, 'Message', Message)

    assert isgoodentity(tp)
    assert tp.parent == tp.get_parent() == dp

    sub = Subscriber(dp)

    assert isgoodentity(sub)
    assert sub.parent == sub.get_parent() == dp

    pub = Publisher(dp)

    assert isgoodentity(pub)
    assert pub.parent == pub.get_parent() == dp

    dr = DataReader(sub, tp)

    assert isgoodentity(dr)
    assert dr.parent == dr.get_parent() == sub

    dw = DataWriter(pub, tp)

    assert isgoodentity(dw)
    assert dw.parent == dw.get_parent() == pub


def test_get_participant():
    dp = DomainParticipant(0)

    assert dp.participant == dp.get_participant() == dp

    tp = Topic(dp, "Message", Message)

    assert isgoodentity(tp)
    assert tp.participant == tp.get_participant() == dp

    sub = Subscriber(dp)

    assert isgoodentity(sub)
    assert sub.participant == sub.get_participant() == dp

    pub = Publisher(dp)

    assert isgoodentity(pub)
    assert pub.participant == pub.get_participant() == dp

    dr = DataReader(sub, tp)

    assert isgoodentity(dr)
    assert dr.participant == dr.get_participant() == dp

    dw = DataWriter(pub, tp)

    assert isgoodentity(dw)
    assert dw.participant == dw.get_participant() == dp


def test_get_children():
    dp = DomainParticipant(0)

    assert len(dp.children) == len(dp.get_children()) == 0

    tp = Topic(dp, "Message", Message)

    assert isgoodentity(tp)
    assert len(dp.children) == len(dp.get_children()) == 1
    assert dp.children[0] == dp.get_children()[0] == tp
    assert len(tp.children) == len(tp.get_children()) == 0

    sub = Subscriber(dp)

    assert isgoodentity(sub)
    assert len(dp.children) == len(dp.get_children()) == 2
    assert set(dp.children) == set([sub, tp])
    assert len(sub.children) == len(sub.get_children()) == 0

    pub = Publisher(dp)

    assert isgoodentity(pub)
    assert len(dp.children) == len(dp.get_children()) == 3
    assert set(dp.children) == set([pub, sub, tp])
    assert len(pub.children) == len(pub.get_children()) == 0

    dr = DataReader(sub, tp)

    assert isgoodentity(dr)
    assert set(dp.children) == set([pub, sub, tp])
    assert len(sub.children) == 1
    assert sub.children[0] == dr

    dw = DataWriter(pub, tp)

    assert isgoodentity(dw)
    assert set(dp.children) == set([pub, sub, tp])
    assert len(pub.children) == 1
    assert pub.children[0] == dw

    del dw
    del dr
    del pub
    del sub
    del tp

    assert len(dp.children) == len(dp.get_children()) == 0


def test_get_pubsub():
    dp = DomainParticipant(0)
    tp = Topic(dp, "Message", Message)
    sub = Subscriber(dp)
    pub = Publisher(dp)
    dr = DataReader(sub, tp)
    dw = DataWriter(pub, tp)

    assert dr.subscriber == dr.get_subscriber() == sub
    assert dw.publisher == dw.get_publisher() == pub

    with pytest.raises(DDSException) as exc:
        dp.get_subscriber()

    assert exc.value.code == DDSException.DDS_RETCODE_ILLEGAL_OPERATION


def test_get_listener():
    dp = DomainParticipant(0, listener=Listener())

    assert dp.get_listener()


def test_get_domain_id():
    dp = DomainParticipant(2)
    assert dp.domain_id == dp.get_domain_id() == 2


def test_get_qos():
    dp = DomainParticipant(0)
    qos = dp.get_qos()
    assert qos == dp.get_qos()


def test_set_listener():
    dp = DomainParticipant(0)
    dp.set_listener(Listener())


def test_listener_reassignment(manual_setup, hitpoint_factory):
    class L(Listener):
        def __init__(self):
            super().__init__()
            self.hitpoint = hitpoint_factory()

        def on_data_available(self, _):
            self.hitpoint.hit()

    l_1 = L()

    # Check we can recieve events
    dr = manual_setup.dr(listener=l_1)
    manual_setup.dw().write(manual_setup.msg)

    assert l_1.hitpoint.was_hit()
    l_1.hitpoint.hp.clear()

    # Check we stop recieving events after setting None
    dr.set_listener(None)
    manual_setup.dw().write(manual_setup.msg)

    assert l_1.hitpoint.was_not_hit()

    # Check we can recieve events after re-assigning the same handler from None
    dr.set_listener(l_1)
    manual_setup.dw().write(manual_setup.msg)

    assert l_1.hitpoint.was_hit()
    l_1.hitpoint.hp.clear()

    # Check we can recieve events from a new handler that replaces the old one
    l_2 = L()
    dr.set_listener(l_2)
    manual_setup.dw().write(manual_setup.msg)

    assert l_1.hitpoint.was_not_hit()
    assert l_2.hitpoint.was_hit()


def test_retain_listener():
    m = lambda x, y: 0
    l = Listener(on_data_available=m)
    dp = DomainParticipant(0, listener=l)
    l = dp.get_listener()
    assert l.on_data_available == m


def test_get_guid():
    dp = DomainParticipant(0)
    dpa = DomainParticipant(1)

    assert dp.guid == dp.get_guid()
    assert dp.guid != dpa.guid


"""
 * Copyright(c) 2021 to 2022 ZettaScale Technology and others
 *
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License v. 2.0 which is available at
 * http://www.eclipse.org/legal/epl-2.0, or the Eclipse Distribution License
 * v. 1.0 which is available at
 * http://www.eclipse.org/org/documents/edl-v10.php.
 *
 * SPDX-License-Identifier: EPL-2.0 OR BSD-3-Clause
"""

import warnings
from dataclasses import dataclass

from .datastruct import datatypes, Integer, String, IntArray, StrArray, IntSequence, StrSequence
from .check_entity_qos import warning_msg

from cyclonedds.pub import Publisher, DataWriter
from cyclonedds.sub import Subscriber, DataReader
from cyclonedds.builtin import BuiltinTopicDcpsPublication, BuiltinTopicDcpsSubscription, BuiltinDataReader
from cyclonedds.topic import Topic
from cyclonedds.dynamic import get_types_for_typeid
from cyclonedds.core import Listener, DDSException, ReadCondition, ViewState, InstanceState, SampleState, WaitSet
from cyclonedds.util import duration


warnings.formatwarning = warning_msg


def discover_datatype(participant, topic_name: str):
    drp = BuiltinDataReader(participant, BuiltinTopicDcpsPublication)
    drpc = ReadCondition(drp, InstanceState.Alive)
    drs = BuiltinDataReader(participant, BuiltinTopicDcpsSubscription)
    drsc = ReadCondition(drs, InstanceState.Alive)
    ws = WaitSet(participant)
    ws.attach(drp)
    ws.attach(drs)

    while True:
        for s in drp.take_iter(condition=drpc, timeout=duration(milliseconds=1)):
            if s.topic_name == topic_name and s.type_id is not None:
                return get_types_for_typeid(participant, s.type_id, duration(seconds=5))
        for s in drs.take_iter(condition=drsc, timeout=duration(milliseconds=1)):
            if s.topic_name == topic_name and s.type_id is not None:
                return get_types_for_typeid(participant, s.type_id, duration(seconds=5))

        ws.wait(duration(milliseconds=4))


class IncompatibleQosWarning(UserWarning):
    pass


class QosListener(Listener):
    def on_requested_incompatible_qos(self, reader, status):
        warnings.warn(
            "The Qos requested for subscription is incompatible with the Qos offered by publication. "
            "PubSub may not be available.",
            IncompatibleQosWarning
        )


@dataclass
class TypeEntities:
    reader: DataReader
    writer: DataWriter


class TopicManager():
    def __init__(self, args, dp, eqos, waitset):
        self.dp = dp

        if args.topic:
            self.topic_name = args.topic
            self.dynamic = False
        else:
            self.topic_name = args.dynamic
            self.dynamic = True

        self.seq = -1  # Sequence number counter
        self.eqos = eqos  # Entity qos
        self.entities = {}  # Store writers and readers
        self.file = args.filename  # Write to file or not
        self.track_samples = {}  # Track read samples if needs to write to file
        try:
            self.listener = QosListener()
            self.pub = Publisher(dp, qos=self.eqos.publisher_qos)
            self.sub = Subscriber(dp, qos=self.eqos.subscriber_qos, listener=self.listener)

            if self.dynamic:
                ds, tmap = discover_datatype(self.dp, self.topic_name)
                self.entity = self.create_entities(ds, self.topic_name)
                self.type_map = tmap
                print(f"Discovered datatype dynamically:{ds}")
            else:
                for type in datatypes:
                    self.entities[type] = self.create_entities(type, self.topic_name + type.postfix())

        except DDSException:
            raise Exception("The arguments inputted are considered invalid for cyclonedds.")

        self.read_cond = ReadCondition((self.entities[Integer] if not self.dynamic else self.entity).reader,
                                       ViewState.Any | InstanceState.Alive | SampleState.NotRead)
        waitset.attach(self.read_cond)

    def write(self, text):
        self.seq += 1
        if self.dynamic:
            # Write dynamic datatype
            try:
                instance = eval(text, globals(), self.type_map)
            except:
                print("Failed to evaluate datatype")
                return
            try:
                self.entity.writer.write(instance)
            except:
                print("Failed to write instance")
            return

        # Write integer
        if type(text) is int:
            self.entities[Integer].writer.write(Integer(self.seq, text))
        elif type(text) is list:
            for i in text:
                if not isinstance(i, type(text[0])):  # Check if elements in the list are the same type
                    raise Exception(
                        "TypeError: Element type inconsistent, "
                        "input list should be a list of integer or a list of string."
                    )

            # Write array or sequence of integer
            if text == [] or isinstance(text[0], int):
                if len(text) == IntArray.size():
                    self.entities[IntArray].writer.write(IntArray(self.seq, text))
                else:
                    self.entities[IntSequence].writer.write(IntSequence(self.seq, text))
            # Write array or sequence of string
            else:
                if len(text) == StrArray.size():
                    self.entities[StrArray].writer.write(StrArray(self.seq, text))
                else:
                    self.entities[StrSequence].writer.write(StrSequence(self.seq, text))
        # Write string
        else:
            self.entities[String].writer.write(String(self.seq, text))

    def read(self):
        if self.dynamic:
            for sample in self.entity.reader.take(N=100):
                print(f"Subscribed: {sample}")
            return
        for type, entity in self.entities.items():
            for sample in entity.reader.take(N=100):
                print(f"Subscribed: {sample}")

                # Track sample to write to file
                if self.file:
                    self.track_samples["sequence " + str(sample.seq)] = {
                        "type": type.postfix(),
                        "keyval": sample.keyval
                    }

    # Create topic, datawriter and datareader
    def create_entities(self, type, topic_name):
        topic = Topic(self.dp, topic_name, type, qos=self.eqos.topic_qos)
        writer = DataWriter(self.pub, topic, qos=self.eqos.datawriter_qos)
        reader = DataReader(self.sub, topic, qos=self.eqos.datareader_qos)
        return TypeEntities(writer=writer, reader=reader)



import pytest

from support_modules.testtopics import Message


def test_regression_write_read_take(common_setup):
    """
        This popped up in early testing, if the refcounting in the C backend is not done right
        the count can become zero between read and take and python frees it.
    """

    common_setup.dw.write(Message(message="Hi!"))
    assert common_setup.dr.read()[0] == Message(message="Hi!")
    assert common_setup.dr.take()[0] == Message(message="Hi!")



"""
 * Copyright(c) 2021 to 2022 ZettaScale Technology and others
 *
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License v. 2.0 which is available at
 * http://www.eclipse.org/legal/epl-2.0, or the Eclipse Distribution License
 * v. 1.0 which is available at
 * http://www.eclipse.org/org/documents/edl-v10.php.
 *
 * SPDX-License-Identifier: EPL-2.0 OR BSD-3-Clause
"""

from mypy.plugin import Plugin, AnalyzeTypeContext
from mypy.types import RawExpressionType, AnyType, TypeOfAny
from mypy.typeanal import make_optional_type


class IdlMyPyPlugin(Plugin):
    def get_type_analyze_hook(self, fullname: str):
        if fullname.startswith("cyclonedds.idl.types"):
            if fullname == "cyclonedds.idl.types.array":
                return strip_length_array_type
            elif fullname == "cyclonedds.idl.types.sequence":
                return strip_length_sequence_type
            elif fullname == "cyclonedds.idl.types.typedef":
                return resolve_type_def
            elif fullname == "cyclonedds.idl.types.bounded_str":
                return bounded_str_to_str
            elif fullname == "cyclonedds.idl.types.default":
                return union_default_type
            elif fullname == "cyclonedds.idl.types.case":
                return union_case_type
        return None

    def get_class_decorator_hook(self, fullname: str):
        from mypy.plugins import dataclasses

        if fullname in ["cyclonedds.idl.IdlStruct", "cyclonedds.idl.IdlUnion"]:
            return dataclasses.dataclass_class_maker_callback


def plugin(version: str):
    return IdlMyPyPlugin


def strip_length_array_type(ctx: AnalyzeTypeContext):
    if len(ctx.type.args) == 0:
        # ref of type itself
        return AnyType(TypeOfAny.special_form)

    if len(ctx.type.args) != 2 or not isinstance(ctx.type.args[1], RawExpressionType) or \
            ctx.type.args[1].base_type_name != "builtins.int":
        ctx.api.fail("cyclonedds.idl.types.array requires two arguments, a subtype and a fixed size.", ctx.context)
        return AnyType(TypeOfAny.from_error)
    return ctx.api.named_type('typing.Sequence', [ctx.api.analyze_type(ctx.type.args[0])])


def resolve_type_def(ctx: AnalyzeTypeContext):
    if len(ctx.type.args) == 0:
        # ref of type itself
        return AnyType(TypeOfAny.special_form)

    if len(ctx.type.args) != 2:
        ctx.api.fail("cyclonedds.idl.types.typedef requires a name and one type argument.", ctx.context)
        return AnyType(TypeOfAny.from_error)
    return ctx.api.analyze_type(ctx.type.args[1])


def strip_length_sequence_type(ctx: AnalyzeTypeContext):
    if len(ctx.type.args) == 0:
        # ref of type itself
        return AnyType(TypeOfAny.special_form)

    if len(ctx.type.args) not in [1, 2]:
        ctx.api.fail("cyclonedds.idl.types.sequence requires a subtype and an optional max size.", ctx.context)
        return AnyType(TypeOfAny.from_error)
    elif len(ctx.type.args) == 2 and not isinstance(ctx.type.args[1], RawExpressionType):
        ctx.api.fail("cyclonedds.idl.types.sequence max size should be an integer.", ctx.context)
        return AnyType(TypeOfAny.from_error)
    elif len(ctx.type.args) == 2 and ctx.type.args[1].base_type_name != "builtins.int":
        ctx.api.fail("cyclonedds.idl.types.sequence max size should be an integer.", ctx.context)
        return AnyType(TypeOfAny.from_error)

    return ctx.api.named_type('typing.Sequence', [ctx.api.analyze_type(ctx.type.args[0])])


def bounded_str_to_str(ctx: AnalyzeTypeContext):
    if len(ctx.type.args) == 0:
        # ref of type itself
        return AnyType(TypeOfAny.special_form)

    if len(ctx.type.args) != 1 or not isinstance(ctx.type.args[0], RawExpressionType) or \
            ctx.type.args[0].base_type_name != "builtins.int":
        ctx.api.fail("cyclonedds.idl.types.bound_str requires one argument, a fixed size.", ctx.context)
        return AnyType(TypeOfAny.from_error)
    return ctx.api.named_type('builtins.str', [])


def union_default_type(ctx: AnalyzeTypeContext):
    if len(ctx.type.args) == 0:
        # ref of type itself
        return AnyType(TypeOfAny.special_form)

    if len(ctx.type.args) != 1:
        ctx.api.fail("cyclonedds.idl.types.default requires one argument, a type.", ctx.context)
        return AnyType(TypeOfAny.from_error)
    return make_optional_type(ctx.api.analyze_type(ctx.type.args[0]))


def union_case_type(ctx: AnalyzeTypeContext):
    if len(ctx.type.args) == 0:
        # ref of type itself
        return AnyType(TypeOfAny.special_form)

    if len(ctx.type.args) != 2:
        ctx.api.fail("cyclonedds.idl.types.case requires two arguments, a discriminator label and a type.", ctx.context)
        return AnyType(TypeOfAny.from_error)
    return make_optional_type(ctx.api.analyze_type(ctx.type.args[1]))



import pytest
import support_modules.test_classes as tc


single_test_data = [
    (tc.SingleInt, (1, 1000, 9128919)),
    (tc.SingleString, ("", "Hello, World!", "aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa")),
    (tc.SingleFloat, (0.0, -1000000000, 1.02)),
    (tc.SingleBool, (True, False)),
    (tc.SingleSequence, ([], [1,2,3], [0, 1] * 100)),
    (tc.SingleArray, ([0, 1, 2],)),
    (tc.SingleUint16, (0, 65535)),
    (tc.SingleBoundedSequence, ([], [1], [1,1], [100, 1, 1])),
    (tc.SingleBoundedString, ("123456789", "llsllë", "")),
    (tc.SingleEnum, (tc.BasicEnum.One, tc.BasicEnum.Two, tc.BasicEnum.Three)),
    (tc.SingleNested, (tc.SingleInt(1),))
]


@pytest.mark.parametrize("_type,values", single_test_data)
def test_simple_datatypes(_type, values):
    for value in values:
        v1 = _type(value=value)
        b = v1.serialize()
        v2 = _type.deserialize(b)
        assert v1 == v2
        assert _type.__idl__.serialize_key_normalized(v1) == _type.__idl__.serialize_key_normalized(v2)


def test_all_primitives():
    v1 = tc.AllPrimitives()
    b = v1.serialize()
    v2 = tc.AllPrimitives.deserialize(b)
    assert v1 == v2


def test_keyed():
    v1 = tc.Keyed(a=1, b=2)
    b = v1.serialize()
    v2 = tc.Keyed.deserialize(b)
    assert v1 == v2
    assert tc.Keyed.__idl__.serialize_key_normalized(v1) == tc.Keyed.__idl__.serialize_key_normalized(v2)
    assert tc.Keyed.__idl__.serialize_key(v1) == bytes.fromhex('01 00 00 00 00 00 00 00')


def test_keyed2():
    v1 = tc.Keyed2(a=1, b=2)
    b = v1.serialize()
    v2 = tc.Keyed2.deserialize(b)
    assert v1 == v2
    assert tc.Keyed2.__idl__.serialize_key_normalized(v1) == tc.Keyed2.__idl__.serialize_key_normalized(v2)
    assert tc.Keyed.__idl__.serialize_key(v2) == bytes.fromhex('01 00 00 00 00 00 00 00')


def test_keyless():
    v1 = tc.Keyless(a=1, b=2)
    b = v1.serialize()
    v2 = tc.Keyless.deserialize(b)
    assert v1 == v2
    assert tc.Keyless.__idl__.serialize_key_normalized(v1) == tc.Keyless.__idl__.serialize_key_normalized(v2)


def test_simple_union():
    values = [
        tc.SingleUnion(value=tc.EasyUnion(a=1)),
        tc.SingleUnion(value=tc.EasyUnion(a=2)),
        tc.SingleUnion(value=tc.EasyUnion(a=3)),
        tc.SingleUnion(value=tc.EasyUnion(b=True))
    ]
    for v1 in values:
        b = v1.serialize()
        v2 = tc.SingleUnion.deserialize(b)
        assert v1 == v2



import pytest

from cyclonedds.domain import DomainParticipant
from cyclonedds.sub import Subscriber
from cyclonedds.util import isgoodentity


def test_subscriber_initialize():
    dp = DomainParticipant(0)
    sub = Subscriber(dp)

    assert isgoodentity(sub)


def test_subscriber_wrong_usage_errors():
    dp = DomainParticipant(0)

    with pytest.raises(TypeError):
        Subscriber(False)

    with pytest.raises(TypeError):
        Subscriber(dp, qos=False)

    with pytest.raises(TypeError):
        Subscriber(dp, listener=False)


@pytest.mark.xfail
def test_subscriber_notify_readers():
    """This will fail, notify_readers is not yet implemented on the C side."""
    dp = DomainParticipant(0)
    sub = Subscriber(dp)
    sub.notify_readers()




"""
 * Copyright(c) 2021 to 2022 ZettaScale Technology and others
 *
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License v. 2.0 which is available at
 * http://www.eclipse.org/legal/epl-2.0, or the Eclipse Distribution License
 * v. 1.0 which is available at
 * http://www.eclipse.org/org/documents/edl-v10.php.
 *
 * SPDX-License-Identifier: EPL-2.0 OR BSD-3-Clause
"""

from argparse import ArgumentError
import uuid
import asyncio
import concurrent
import ctypes as ct
from weakref import WeakValueDictionary
from typing import Any, Callable, Dict, Optional, List, TYPE_CHECKING
from datetime import datetime, time, timedelta

from .internal import c_call, c_callable, dds_infinity, dds_c_t, DDS, stat_keyvalue, stat_kind
from .qos import Qos, Policy, _CQos


if TYPE_CHECKING:
    import cyclonedds


class DDSException(Exception):
    """This exception is thrown when a return code from the underlying C api indicates non-valid use of the API.
    Print the exception directly or convert it to string for a detailed description.

    Attributes
    ----------
    code: int
        One of the ``DDS_RETCODE_`` constants that indicates the type of error.
    msg: str
        A human readable description of where the error occurred
    """

    DDS_RETCODE_OK = 0  # Success
    DDS_RETCODE_ERROR = -1  # Non specific error
    DDS_RETCODE_UNSUPPORTED = -2  # Feature unsupported
    DDS_RETCODE_BAD_PARAMETER = -3  # Bad parameter value
    DDS_RETCODE_PRECONDITION_NOT_MET = -4  # Precondition for operation not met
    DDS_RETCODE_OUT_OF_RESOURCES = (
        -5
    )  # When an operation fails because of a lack of resources
    DDS_RETCODE_NOT_ENABLED = -6  # When a configurable feature is not enabled
    DDS_RETCODE_IMMUTABLE_POLICY = (
        -7
    )  # When an attempt is made to modify an immutable policy
    DDS_RETCODE_INCONSISTENT_POLICY = (
        -8
    )  # When a policy is used with inconsistent values
    DDS_RETCODE_ALREADY_DELETED = (
        -9
    )  # When an attempt is made to delete something more than once
    DDS_RETCODE_TIMEOUT = -10  # When a timeout has occurred
    DDS_RETCODE_NO_DATA = -11  # When expected data is not provided
    DDS_RETCODE_ILLEGAL_OPERATION = (
        -12
    )  # When a function is called when it should not be
    DDS_RETCODE_NOT_ALLOWED_BY_SECURITY = (
        -13
    )  # When credentials are not enough to use the function

    error_message_mapping = {
        DDS_RETCODE_OK: ("DDS_RETCODE_OK", "Success"),
        DDS_RETCODE_ERROR: ("DDS_RETCODE_ERROR", "Non specific error"),
        DDS_RETCODE_UNSUPPORTED: ("DDS_RETCODE_UNSUPPORTED", "Feature unsupported"),
        DDS_RETCODE_BAD_PARAMETER: ("DDS_RETCODE_BAD_PARAMETER", "Bad parameter value"),
        DDS_RETCODE_PRECONDITION_NOT_MET: (
            "DDS_RETCODE_PRECONDITION_NOT_MET",
            "Precondition for operation not met",
        ),
        DDS_RETCODE_OUT_OF_RESOURCES: (
            "DDS_RETCODE_OUT_OF_RESOURCES",
            "Operation failed because of a lack of resources",
        ),
        DDS_RETCODE_NOT_ENABLED: (
            "DDS_RETCODE_NOT_ENABLED",
            "A configurable feature is not enabled",
        ),
        DDS_RETCODE_IMMUTABLE_POLICY: (
            "DDS_RETCODE_IMMUTABLE_POLICY",
            "An attempt was made to modify an immutable policy",
        ),
        DDS_RETCODE_INCONSISTENT_POLICY: (
            "DDS_RETCODE_INCONSISTENT_POLICY",
            "A policy with inconsistent values was used",
        ),
        DDS_RETCODE_ALREADY_DELETED: (
            "DDS_RETCODE_ALREADY_DELETED",
            "An attempt was made to delete something more than once",
        ),
        DDS_RETCODE_TIMEOUT: ("DDS_RETCODE_TIMEOUT", "A timeout has occurred"),
        DDS_RETCODE_NO_DATA: ("DDS_RETCODE_NO_DATA", "Expected data is not provided"),
        DDS_RETCODE_ILLEGAL_OPERATION: (
            "DDS_RETCODE_ILLEGAL_OPERATION",
            "A function was called when it should not be",
        ),
        DDS_RETCODE_NOT_ALLOWED_BY_SECURITY: (
            "DDS_RETCODE_NOT_ALLOWED_BY_SECURITY",
            "Insufficient credentials supplied to use the function",
        ),
    }

    def __init__(self, code: int, msg: str = None, **kwargs) -> None:
        """Initialize a DDSException. Code should be one of the DDS_RETCODE_* constants."""
        self.code = code
        self.msg = msg or ""
        super().__init__(**kwargs)

    def __str__(self) -> str:
        if self.code in self.error_message_mapping:
            msg = self.error_message_mapping[self.code]
            return f"[{msg[0]}] {msg[1]}. {self.msg}"
        return f"[DDSException] Got an unexpected error code '{self.code}'. {self.msg}"

    def __repr__(self) -> str:
        return str(self)


class Entity(DDS):
    """
    Base class for all entities in the DDS API. The lifetime of the underlying
    DDS API object is linked to the lifetime of the Python entity object.

    Attributes
    ----------
    subscriber
                 If this entity is associated with a DataReader retrieve it.
                 It is read-only. This is a proxy for get_subscriber().
    publisher
                 If this entity is associated with a Publisher retrieve it.
                 It is read-only. This is a proxy for get_publisher().
    datareader
                 If this entity is associated with a DataReader retrieve it.
                 It is read-only. This is a proxy for get_datareader().
    guid:        uuid.UUID
                 Return the globally unique identifier for this entity.
                 It is read-only. This is a proxy for get_guid().
    status_mask
                 The status mask for this entity. It is a set of bits formed
                 from ``DDSStatus``. This is a proxy for get/set_status_mask().
    parent
                 The entity that is this entities parent. For example: the subscriber for a
                 datareader, the participant for a topic.
                 It is read-only. This is a proxy for get_parent().
    participant
                 Get the participant for any entity, will only fail for a ``Domain``.
                 It is read-only. This is a proxy for get_participant().
    children
                 Get a list of children belonging to this entity. It is the opposite as ``parent``.
                 It is read-only. This is a proxy for get_children().
    domain_id
                 Get the id of the domain this entity belongs to.
    """

    _entities: Dict[dds_c_t.entity, "Entity"] = WeakValueDictionary()

    def __init__(self, ref: int, listener: "Listener" = None) -> None:
        """Initialize an Entity. You should never need to initialize an Entity manually.

        Parameters
        ----------
        ref: int
            The reference id as returned by the DDS API.
        listener: Listener
            Listener for this entity. We retain the python object to avoid it being garbage collected if the listener
            goes out of scope but the entity doesn't. If we don't the python function will be freed, causing C to call
            into freed memory -> segfault.

        Raises
        ------
        DDSException
            If an invalid reference id is passed to this function this means instantiation of some other object failed.
        """
        if ref < 0:
            raise DDSException(
                ref,
                f"Occurred upon initialisation of a {self.__class__.__module__}.{self.__class__.__name__}",
            )
        super().__init__(ref)
        self._entities[self._ref] = self
        self._listener = listener

    def __del__(self) -> None:
        if not hasattr(self, "_ref") or self._ref not in self._entities:
            return

        del self._entities[self._ref]
        self._delete(self._ref)

    def get_subscriber(self) -> Optional["cyclonedds.sub.Subscriber"]:
        """Retrieve the subscriber associated with this entity.

        Returns
        -------
        Optional[Subscriber]
            Not all entities are associated with a subscriber, so this method may return None.

        Raises
        ------
        DDSException
        """
        ref = self._get_subscriber(self._ref)
        if ref >= 0:
            return self.get_entity(ref)
        raise DDSException(
            ref, f"Occurred when getting the subscriber for {repr(self)}"
        )

    subscriber: Optional["cyclonedds.sub.Subscriber"] = property(get_subscriber)

    def get_publisher(self) -> Optional["cyclonedds.pub.Publisher"]:
        """Retrieve the publisher associated with this entity.

        Returns
        -------
        Optional[Publisher]
            Not all entities are associated with a publisher, so this method may return None.

        Raises
        ------
        DDSException
        """
        ref = self._get_publisher(self._ref)
        if ref >= 0:
            return self.get_entity(ref)
        raise DDSException(ref, f"Occurred when getting the publisher for {repr(self)}")

    publisher: Optional["cyclonedds.pub.Publisher"] = property(get_publisher)

    def get_datareader(self) -> Optional["cyclonedds.sub.DataReader"]:
        """Retrieve the datareader associated with this entity.

        Returns
        -------
        Optional[DataReader]
            Not all entities are associated with a datareader, so this method may return None.

        Raises
        ------
        DDSException
        """
        ref = self._get_datareader(self._ref)
        if ref >= 0:
            return self.get_entity(ref)
        raise DDSException(
            ref, f"Occurred when getting the datareader for {repr(self)}"
        )

    datareader: Optional["cyclonedds.sub.DataReader"] = property(get_datareader)

    def get_instance_handle(self) -> int:
        """Retrieve the instance associated with this entity.

        Returns
        -------
        int
            The integer handle is just a number you can use in writer/reader calls.

        Raises
        ------
        DDSException
        """
        handle = dds_c_t.instance_handle()
        ret = self._get_instance_handle(self._ref, ct.byref(handle))
        if ret == 0:
            return int(handle.value)
        raise DDSException(
            ret, f"Occurred when getting the instance handle for {repr(self)}"
        )

    instance_handle: int = property(get_instance_handle)

    def get_guid(self) -> uuid.UUID:
        """Get a globally unique identifier for this entity.

        Returns
        -------
        uuid.UUID
            View the python documentation for this class for detailed usage.

        Raises
        ------
        DDSException
        """
        guid = dds_c_t.guid()
        ret = self._get_guid(self._ref, ct.byref(guid))
        if ret == 0:
            return guid.as_python_guid()
        raise DDSException(ret, f"Occurred when getting the GUID for {repr(self)}")

    guid: uuid.UUID = property(get_guid)

    def read_status(self, mask: int = None) -> int:
        """Read the status bits set on this Entity. You can build a mask by using :class:`DDSStatus`.

        Parameters
        ----------
        mask
            The :class:`DDSStatus` mask. If not supplied the mask is used that was set on this Entity using set_status_mask.

        Returns
        -------
        int
            The :class:`DDSStatus` bits that were set.

        Raises
        ------
        DDSException
        """
        status = ct.c_uint32()
        ret = self._read_status(
            self._ref,
            ct.byref(status),
            ct.c_uint32(mask) if mask else self.get_status_mask(),
        )
        if ret == 0:
            return status.value
        raise DDSException(ret, f"Occurred when reading the status for {repr(self)}")

    def take_status(self, mask: int = None) -> int:
        """Take the status bits set on this Entity, after which they will be set to 0 again.
        You can build a mask by using :class:`DDSStatus`.

        Parameters
        ----------
        mask
            The :class:`DDSStatus` mask. If not supplied the mask is used that was set on this Entity using set_status_mask.

        Returns
        -------
        int
            The :class:`DDSStatus` bits that were set.

        Raises
        ------
        DDSException
        """
        status = ct.c_uint32()
        ret = self._take_status(
            self._ref,
            ct.byref(status),
            ct.c_uint32(mask) if mask else self.get_status_mask(),
        )
        if ret == 0:
            return status.value
        raise DDSException(ret, f"Occurred when taking the status for {repr(self)}")

    def get_status_changes(self) -> int:
        """Get all status changes since the last read_status() or take_status().

        Returns
        -------
        int
            The :class:`DDSStatus` bits that were set.

        Raises
        ------
        DDSException
        """
        status = ct.c_uint32()
        ret = self._get_status_changes(self._ref, ct.byref(status))
        if ret == 0:
            return status.value
        raise DDSException(
            ret, f"Occurred when getting the status changes for {repr(self)}"
        )

    def get_status_mask(self) -> int:
        """Get the status mask for this Entity.

        Returns
        -------
        int
            The :class:`DDSStatus` bits that are enabled.

        Raises
        ------
        DDSException
        """
        mask = ct.c_uint32()
        ret = self._get_status_mask(self._ref, ct.byref(mask))
        if ret == 0:
            return mask.value
        raise DDSException(
            ret, f"Occurred when getting the status mask for {repr(self)}"
        )

    def set_status_mask(self, mask: int) -> None:
        """Set the status mask for this Entity. By default the mask is 0. Only the status changes
            for the bits in this mask are tracked on this entity.

        Parameters
        ----------
        mask : int
            The :class:`DDSStatus` bits to track.

        Raises
        ------
        DDSException
        """
        ret = self._set_status_mask(self._ref, ct.c_uint32(mask))
        if ret == 0:
            return
        raise DDSException(
            ret, f"Occurred when setting the status mask for {repr(self)}"
        )

    status_mask: int = property(get_status_mask, set_status_mask)

    def get_qos(self) -> Qos:
        """Get the :class:`Qos` associated with this entity. Note that the object returned is not
        the same python object that you used to set the :class:`Qos` on this object. Modifications to the :class:`Qos` object
        that is returned does **not** modify the Qos of the Entity.

        Returns
        -------
        Qos
            The :class:`Qos` object associated with this entity.

        Raises
        ------
        DDSException
        """
        cqos = _CQos.cqos_create()
        ret = self._get_qos(self._ref, cqos)
        if ret == 0:
            qos = _CQos.cqos_to_qos(cqos)
            _CQos.cqos_destroy(cqos)
            return qos
        _CQos.cqos_destroy(cqos)
        raise DDSException(
            ret, f"Occurred when getting the Qos Policies for {repr(self)}"
        )

    def set_qos(self, qos: Qos) -> None:
        """Set :class:`Qos` policies on this entity. Note, only a limited number of :class:`Qos` policies can be set after
        the object is created (:class:`Policy.LatencyBudget` and :class:`Policy.OwnershipStrength`). Any policies not set
        explicitly in the supplied :class:`Qos` remain unchanged.

        Parameters
        ----------
        qos : Qos
            The :class:`Qos` to apply to this entity.

        Raises
        ------
        DDSException
            If you pass an immutable policy or cause the total collection of qos policies to become inconsistent
            an exception will be raised.
        """
        cqos = _CQos.qos_to_cqos(qos)
        ret = self._set_qos(self._ref, cqos)
        _CQos.cqos_destroy(cqos)
        if ret == 0:
            return
        raise DDSException(
            ret, f"Occurred when setting the Qos Policies for {repr(self)}"
        )

    def get_listener(self) -> "Listener":
        """Return a listener with the right methods set. Modifying the returned listener object does not modify
        this entity, you will have to call set_listener() with the changed object.

        Returns
        -------
        Listener
            A listener with which you can add additional callbacks.
        """
        return self._listener.copy() if self._listener else Listener()

    def set_listener(self, listener: Optional["Listener"]) -> None:
        """Update the listener for this object. If a listener already exist for this object only the fields you explicitly
        have set on your new listener are overwritten. Passing None will remove this entity's Listener.

        Future changes to the passed Listener object will not affect the Listener associated with this Entity.

        Parameters
        ----------
        listener :
            The listener object to use, or None to remove the current listener from this Entity.

        Raises
        ------
        DDSException
        """
        if listener is not None:
            if self._listener is not None:
                if self._listener != listener:
                    listener.copy_to(self._listener)
            else:
                self._listener = listener.copy()
            ref = self._listener._ref
        else:
            ref = None
            self._listener = None

        ret = self._set_listener(self._ref, ref)
        if ret == 0:
            return
        raise DDSException(ret, f"Occurred when setting the Listener for {repr(self)}")

    def get_parent(self) -> Optional["Entity"]:
        """Get the parent entity associated with this entity. A ``Domain`` object is the only object without parent,
        but if the domain is not created through the Python API this call won't find it
        and return None from the DomainParticipant.

        Returns
        -------
        Optional[Entity]
            The parent of this entity. This would be the Subscriber for a DataReader, DomainParticipant for a Topic etc.

        Raises
        ------
        DDSException
        """
        ret = self._get_parent(self._ref)

        if ret > 0:
            return self.get_entity(ret)

        if ret is None or ret == 0:
            return None

        raise DDSException(ret, f"Occurred when getting the parent of {repr(self)}")

    parent: Optional["Entity"] = property(get_parent)

    def get_participant(self) -> Optional["cyclonedds.domain.DomainParticipant"]:
        """Get the domain participant for this entity. This should work on all valid Entity objects except a Domain.

        Returns
        -------
        Optional[cyclonedds.domain.DomainParticipant]
            Only fails for a Domain object.

        Raises
        ------
        DDSException
        """
        ret = self._get_participant(self._ref)

        if ret > 0:
            return self.get_entity(ret)

        if ret is None or ret == 0:
            return None

        raise DDSException(
            ret, f"Occurred when getting the participant of {repr(self)}"
        )

    participant: Optional["cyclonedds.domain.DomainParticipant"] = property(get_participant)

    def get_children(self) -> List["Entity"]:
        """Get the list of children of this entity. For example, the list of datareaders belonging to a subscriber.
        Opposite of parent.

        Returns
        -------
        List[Entity]
            All the entities considered children of this entity.

        Raises
        ------
        DDSException
        """
        num_children = self._get_children(self._ref, None, 0)
        if num_children < 0:
            raise DDSException(
                num_children,
                f"Occurred when getting the number of children of {repr(self)}",
            )
        if num_children == 0:
            return []

        children_list = (dds_c_t.entity * int(num_children))()
        children_list_pt = ct.cast(children_list, ct.POINTER(dds_c_t.entity))

        ret = self._get_children(self._ref, children_list_pt, num_children)
        if ret >= 0:
            return [self.get_entity(children_list[i]) for i in range(ret)]

        raise DDSException(ret, f"Occurred when getting the children of {repr(self)}")

    children: List["Entity"] = property(get_children)

    def get_domain_id(self) -> int:
        """Get the id of the domain this entity resides in.

        Returns
        -------
        int
            The domain id.

        Raises
        ------
        DDSException
        """
        domainid = dds_c_t.domainid()
        ret = self._get_domainid(self._ref, ct.byref(domainid))
        if ret == 0:
            return domainid.value

        raise DDSException(ret, f"Occurred when getting the domainid of {repr(self)}")

    domain_id: int = property(get_domain_id)

    def begin_coherent(self) -> None:
        """Begin coherent publishing or begin accessing a coherent set in a Subscriber.

        This can only be invoked on Publishers, Subscribers, DataWriters and DataReaders.
        Invoking on a DataWriter or DataReader behaves as if it was invoked on its parent
        Publisher or Subscriber respectively.
        """
        ret = self._begin_coherent(self._ref)
        if ret < 0:
            raise DDSException(ret, f"Occurred when beginning coherent on {repr(self)}")

    def end_coherent(self) -> None:
        """End coherent publishing or end accessing a coherent set in a Subscriber.

        This can only be invoked on Publishers, Subscribers, DataWriters and DataReaders.
        Invoking on a DataWriter or DataReader behaves as if it was invoked on its parent
        Publisher or Subscriber respectively.
        """
        ret = self._end_coherent(self._ref)
        if ret < 0:
            raise DDSException(ret, f"Occurred when ending coherent on {repr(self)}")

    @classmethod
    def get_entity(cls, entity_id) -> Optional["Entity"]:
        """Turn a CycloneDDS C Entity id into a Python object. You shouldn't need this."""
        return cls._entities.get(entity_id)

    @classmethod
    def _init_from_retcode(cls, code):
        entity = cls.__new__(cls)
        Entity.__init__(entity, code)
        return entity

    @c_call("dds_delete")
    def _delete(self, entity: dds_c_t.entity) -> None:
        pass

    @c_call("dds_get_subscriber")
    def _get_subscriber(self, entity: dds_c_t.entity) -> dds_c_t.entity:
        pass

    @c_call("dds_get_datareader")
    def _get_datareader(self, entity: dds_c_t.entity) -> dds_c_t.entity:
        pass

    @c_call("dds_get_publisher")
    def _get_publisher(self, entity: dds_c_t.entity) -> dds_c_t.entity:
        pass

    @c_call("dds_get_instance_handle")
    def _get_instance_handle(
        self, entity: dds_c_t.entity, handle: ct.POINTER(dds_c_t.instance_handle)
    ) -> dds_c_t.returnv:
        pass

    @c_call("dds_get_guid")
    def _get_guid(
        self, entity: dds_c_t.entity, guid: ct.POINTER(dds_c_t.guid)
    ) -> dds_c_t.returnv:
        pass

    @c_call("dds_read_status")
    def _read_status(
        self, entity: dds_c_t.entity, status: ct.POINTER(ct.c_uint32), mask: ct.c_uint32
    ) -> dds_c_t.returnv:
        pass

    @c_call("dds_take_status")
    def _take_status(
        self, entity: dds_c_t.entity, status: ct.POINTER(ct.c_uint32), mask: ct.c_uint32
    ) -> dds_c_t.returnv:
        pass

    @c_call("dds_get_status_changes")
    def _get_status_changes(
        self, entity: dds_c_t.entity, status: ct.POINTER(ct.c_uint32)
    ) -> dds_c_t.returnv:
        pass

    @c_call("dds_get_status_mask")
    def _get_status_mask(
        self, entity: dds_c_t.entity, mask: ct.POINTER(ct.c_uint32)
    ) -> dds_c_t.returnv:
        pass

    @c_call("dds_set_status_mask")
    def _set_status_mask(
        self, entity: dds_c_t.entity, mask: ct.c_uint32
    ) -> dds_c_t.returnv:
        pass

    @c_call("dds_get_qos")
    def _get_qos(self, entity: dds_c_t.entity, qos: dds_c_t.qos_p) -> dds_c_t.returnv:
        pass

    @c_call("dds_set_qos")
    def _set_qos(self, entity: dds_c_t.entity, qos: dds_c_t.qos_p) -> dds_c_t.returnv:
        pass

    @c_call("dds_get_listener")
    def _get_listener(
        self, entity: dds_c_t.entity, listener: dds_c_t.listener_p
    ) -> dds_c_t.returnv:
        pass

    @c_call("dds_set_listener")
    def _set_listener(
        self, entity: dds_c_t.entity, listener: dds_c_t.listener_p
    ) -> dds_c_t.returnv:
        pass

    @c_call("dds_get_parent")
    def _get_parent(self, entity: dds_c_t.entity) -> dds_c_t.entity:
        pass

    @c_call("dds_get_participant")
    def _get_participant(self, entity: dds_c_t.entity) -> dds_c_t.entity:
        pass

    @c_call("dds_get_children")
    def _get_children(
        self,
        entity: dds_c_t.entity,
        children_list: ct.POINTER(dds_c_t.returnv),
        size: ct.c_size_t,
    ) -> dds_c_t.returnv:
        pass

    @c_call("dds_get_domainid")
    def _get_domainid(
        self, entity: dds_c_t.entity, domainid: ct.POINTER(dds_c_t.domainid)
    ) -> dds_c_t.returnv:
        pass

    @c_call("dds_begin_coherent")
    def _begin_coherent(self, entity: dds_c_t.entity) -> dds_c_t.returnv:
        pass

    @c_call("dds_end_coherent")
    def _end_coherent(self, entity: dds_c_t.entity) -> dds_c_t.returnv:
        pass

    def __repr__(self) -> str:
        ref = None
        try:
            ref = self._ref
        except Exception:
            pass
        return f"<Entity, type={self.__class__.__module__}.{self.__class__.__name__}, addr={hex(id(self))}, id={ref}>"


_inconsistent_topic_fn = c_callable(
    None, [dds_c_t.entity, dds_c_t.inconsistent_topic_status, ct.c_void_p]
)
_data_available_fn = c_callable(None, [dds_c_t.entity, ct.c_void_p])
_liveliness_lost_fn = c_callable(
    None, [dds_c_t.entity, dds_c_t.liveliness_lost_status, ct.c_void_p]
)
_liveliness_changed_fn = c_callable(
    None, [dds_c_t.entity, dds_c_t.liveliness_changed_status, ct.c_void_p]
)
_offered_deadline_missed_fn = c_callable(
    None, [dds_c_t.entity, dds_c_t.offered_deadline_missed_status, ct.c_void_p]
)
_offered_incompatible_qos_fn = c_callable(
    None, [dds_c_t.entity, dds_c_t.offered_incompatible_qos_status, ct.c_void_p]
)
_data_on_readers_fn = c_callable(None, [dds_c_t.entity, ct.c_void_p])
_on_sample_lost_fn = c_callable(
    None, [dds_c_t.entity, dds_c_t.sample_lost_status, ct.c_void_p]
)
_on_sample_rejected_fn = c_callable(
    None, [dds_c_t.entity, dds_c_t.sample_rejected_status, ct.c_void_p]
)
_on_requested_deadline_missed_fn = c_callable(
    None, [dds_c_t.entity, dds_c_t.requested_deadline_missed_status, ct.c_void_p]
)
_on_requested_incompatible_qos_fn = c_callable(
    None, [dds_c_t.entity, dds_c_t.requested_incompatible_qos_status, ct.c_void_p]
)
_on_publication_matched_fn = c_callable(
    None, [dds_c_t.entity, dds_c_t.publication_matched_status, ct.c_void_p]
)
_on_subscription_matched_fn = c_callable(
    None, [dds_c_t.entity, dds_c_t.subscription_matched_status, ct.c_void_p]
)


def _is_override(func):
    obj = func.__self__
    if type(obj) == Listener:
        return False

    parent_method = getattr(super(type(obj), obj), func.__name__)
    return func.__func__ != parent_method.__func__


class Listener(DDS):
    """Listeners are callback containers for entities."""

    def __init__(self, **kwargs):
        """Create a Listener object. The initializer takes override function lambdas.

        Please note that all listener callbacks are dispatched synchronously from DDS receive thread(s). You can get away with doing
        tiny amounts of processing in these callback methods, but aquiring the Python GIL from the DDS receive thread will severely
        hurt your DDS performance. Furthermore, deleting entities or writing data inside listener callbacks can get you into deadlocks.

        Parameters
        ----------
        on_data_available : Callable
            Set on_data_available callback.
        on_inconsistent_topic : Callable
            Set on_inconsistent_topic callback.
        on_liveliness_lost : Callable
            Set on_liveliness_lost callback.
        on_liveliness_changed : Callable
            Set on_liveliness_changed callback.
        on_offered_deadline_missed : Callable
            Set on_offered_deadline_missed callback.
        on_offered_incompatible_qos : Callable
            Set on_offered_incompatible_qos callback.
        on_data_on_readers : Callable
            Set on_data_on_readers callback.
        on_sample_lost : Callable
            Set on_sample_lost callback.
        on_sample_rejected : Callable
            Set on_sample_rejected callback.
        on_requested_deadline_missed : Callable
            Set on_requested_deadline_missed callback.
        on_requested_incompatible_qos : Callable
            Set on_requested_incompatible_qos callback.
        on_publication_matched : Callable
            Set on_publication_matched callback.
        on_subscription_matched : Callable
            Set on_subscription_matched callback.
        """
        super().__init__(self._create_listener(None))
        self._set_functors = {}

        if _is_override(self.on_data_available):
            self.set_on_data_available(self.on_data_available)
            self._set_functors["on_data_available"] = self.on_data_available

        if _is_override(self.on_inconsistent_topic):
            self.set_on_inconsistent_topic(self.on_inconsistent_topic)
            self._set_functors["on_inconsistent_topic"] = self.on_inconsistent_topic

        if _is_override(self.on_liveliness_lost):
            self.set_on_liveliness_lost(self.on_liveliness_lost)
            self._set_functors["on_liveliness_lost"] = self.on_liveliness_lost

        if _is_override(self.on_liveliness_changed):
            self.set_on_liveliness_changed(self.on_liveliness_changed)
            self._set_functors["on_liveliness_changed"] = self.on_liveliness_changed

        if _is_override(self.on_offered_deadline_missed):
            self.set_on_offered_deadline_missed(self.on_offered_deadline_missed)
            self._set_functors[
                "on_offered_deadline_missed"
            ] = self.on_offered_deadline_missed

        if _is_override(self.on_offered_incompatible_qos):
            self.set_on_offered_incompatible_qos(self.on_offered_incompatible_qos)
            self._set_functors[
                "on_offered_incompatible_qos"
            ] = self.on_offered_incompatible_qos

        if _is_override(self.on_data_on_readers):
            self.set_on_data_on_readers(self.on_data_on_readers)
            self._set_functors["on_data_on_readers"] = self.on_data_on_readers

        if _is_override(self.on_sample_lost):
            self.set_on_sample_lost(self.on_sample_lost)
            self._set_functors["on_sample_lost"] = self.on_sample_lost

        if _is_override(self.on_sample_rejected):
            self.set_on_sample_rejected(self.on_sample_rejected)
            self._set_functors["on_sample_rejected"] = self.on_sample_rejected

        if _is_override(self.on_requested_deadline_missed):
            self.set_on_requested_deadline_missed(self.on_requested_deadline_missed)
            self._set_functors[
                "on_requested_deadline_missed"
            ] = self.on_requested_deadline_missed

        if _is_override(self.on_requested_incompatible_qos):
            self.set_on_requested_incompatible_qos(self.on_requested_incompatible_qos)
            self._set_functors[
                "on_requested_incompatible_qos"
            ] = self.on_requested_incompatible_qos

        if _is_override(self.on_publication_matched):
            self.set_on_publication_matched(self.on_publication_matched)
            self._set_functors["on_publication_matched"] = self.on_publication_matched

        if _is_override(self.on_subscription_matched):
            self.set_on_subscription_matched(self.on_subscription_matched)
            self._set_functors["on_subscription_matched"] = self.on_subscription_matched

        self.setters = {
            "on_data_available": self.set_on_data_available,
            "on_inconsistent_topic": self.set_on_inconsistent_topic,
            "on_liveliness_lost": self.set_on_liveliness_lost,
            "on_liveliness_changed": self.set_on_liveliness_changed,
            "on_offered_deadline_missed": self.set_on_offered_deadline_missed,
            "on_offered_incompatible_qos": self.set_on_offered_incompatible_qos,
            "on_data_on_readers": self.set_on_data_on_readers,
            "on_sample_lost": self.set_on_sample_lost,
            "on_sample_rejected": self.set_on_sample_rejected,
            "on_requested_deadline_missed": self.set_on_requested_deadline_missed,
            "on_requested_incompatible_qos": self.set_on_requested_incompatible_qos,
            "on_publication_matched": self.set_on_publication_matched,
            "on_subscription_matched": self.set_on_subscription_matched,
        }

        for name, value in kwargs.items():
            if name not in self.setters:
                raise ArgumentError(f"Invalid listener attribute '{name}'")
            self.setters[name](value)

    def __del__(self):
        self._delete_listener(self._ref)

    def reset(self) -> None:
        self._reset_listener(self._ref)

    def copy(self) -> "Listener":
        listener = Listener(**self._set_functors)
        return listener

    def copy_to(self, listener: "Listener") -> None:
        for name, functor in self._set_functors.items():
            listener.setters[name](functor)

    def merge(self, listener: "Listener") -> None:
        """
        Copies any configured (non-default) callbacks from the given `listener` to self, replacing existing callbacks
        already configured on this listener.
        """
        listener.copy_to(self)

    def on_inconsistent_topic(
        self,
        reader: "cyclonedds.sub.DataReader",
        status: dds_c_t.inconsistent_topic_status,
    ) -> None:
        pass

    def set_on_inconsistent_topic(
        self, callable: Callable[["cyclonedds.sub.DataReader"], None]
    ):
        self.on_inconsistent_topic = callable
        if callable is None:
            self._set_inconsistent_topic(self._ref, None)
            del self._set_functors["on_inconsistent_topic"]
        else:
            self._set_functors["on_inconsistent_topic"] = self.on_inconsistent_topic

            def call(topic, status, arg):
                self.on_inconsistent_topic(Entity.get_entity(topic), status)

            self._on_inconsistent_topic = _inconsistent_topic_fn(call)
            self._set_inconsistent_topic(self._ref, self._on_inconsistent_topic)

    def on_data_available(self, reader: "cyclonedds.sub.DataReader") -> None:
        pass

    def set_on_data_available(
        self, callable: Callable[["cyclonedds.sub.DataReader"], None]
    ):
        self.on_data_available = callable
        if callable is None:
            self._set_data_available(self._ref, None)
            del self._set_functors["on_data_available"]
        else:
            self._set_functors["on_data_available"] = self.on_data_available

            def call(reader, arg):
                self.on_data_available(Entity.get_entity(reader))

            self._on_data_available = _data_available_fn(call)
            self._set_data_available(self._ref, self._on_data_available)

    def on_liveliness_lost(
        self,
        writer: "cyclonedds.pub.DataWriter",
        status: dds_c_t.liveliness_lost_status,
    ) -> None:
        pass

    def set_on_liveliness_lost(
        self,
        callable: Callable[
            ["cyclonedds.pub.DataWriter", dds_c_t.liveliness_lost_status], None
        ],
    ):
        self.on_liveliness_lost = callable
        if callable is None:
            self._set_liveliness_lost(self._ref, None)
            del self._set_functors["on_liveliness_lost"]
        else:
            self._set_functors["on_liveliness_lost"] = self.on_liveliness_lost

            def call(writer, status, arg):
                self.on_liveliness_lost(Entity.get_entity(writer), status)

            self._on_liveliness_lost = _liveliness_lost_fn(call)
            self._set_liveliness_lost(self._ref, self._on_liveliness_lost)

    def on_liveliness_changed(
        self,
        reader: "cyclonedds.sub.DataReader",
        status: dds_c_t.liveliness_changed_status,
    ) -> None:
        pass

    def set_on_liveliness_changed(
        self,
        callable: Callable[
            ["cyclonedds.sub.DataReader", dds_c_t.liveliness_changed_status], None
        ],
    ):
        self.on_liveliness_changed = callable
        if callable is None:
            self._set_liveliness_changed(self._ref, None)
            del self._set_functors["on_liveliness_changed"]
        else:
            self._set_functors["on_liveliness_changed"] = self.on_liveliness_changed

            def call(reader, status, arg):
                self.on_liveliness_changed(Entity.get_entity(reader), status)

            self._on_liveliness_changed = _liveliness_changed_fn(call)
            self._set_liveliness_changed(self._ref, self._on_liveliness_changed)

    def on_offered_deadline_missed(
        self,
        writer: "cyclonedds.pub.DataWriter",
        status: dds_c_t.offered_deadline_missed_status,
    ) -> None:
        pass

    def set_on_offered_deadline_missed(
        self,
        callable: Callable[
            ["cyclonedds.pub.DataWriter", dds_c_t.offered_deadline_missed_status], None
        ],
    ):
        self.on_offered_deadline_missed = callable
        if callable is None:
            self._set_on_offered_deadline_missed(self._ref, None)
            del self._set_functors["on_offered_deadline_missed"]
        else:
            self._set_functors[
                "on_offered_deadline_missed"
            ] = self.on_offered_deadline_missed

            def call(writer, status, arg):
                self.on_offered_deadline_missed(Entity.get_entity(writer), status)

            self._on_offered_deadline_missed = _offered_deadline_missed_fn(call)
            self._set_on_offered_deadline_missed(
                self._ref, self._on_offered_deadline_missed
            )

    def on_offered_incompatible_qos(
        self,
        writer: "cyclonedds.pub.DataWriter",
        status: dds_c_t.offered_incompatible_qos_status,
    ) -> None:
        pass

    def set_on_offered_incompatible_qos(
        self,
        callable: Callable[
            ["cyclonedds.pub.DataWriter", dds_c_t.offered_incompatible_qos_status], None
        ],
    ):
        self.on_offered_incompatible_qos = callable
        if callable is None:
            self._set_on_offered_incompatible_qos(self._ref, None)
            del self._set_functors["on_offered_incompatible_qos"]
        else:
            self._set_functors[
                "on_offered_incompatible_qos"
            ] = self.on_offered_incompatible_qos

            def call(writer, status, arg):
                self.on_offered_incompatible_qos(Entity.get_entity(writer), status)

            self._on_offered_incompatible_qos = _offered_incompatible_qos_fn(call)
            self._set_on_offered_incompatible_qos(
                self._ref, self._on_offered_incompatible_qos
            )

    def on_data_on_readers(self, subscriber: "cyclonedds.sub.Subscriber") -> None:
        pass

    def set_on_data_on_readers(
        self, callable: Callable[["cyclonedds.sub.Subscriber"], None]
    ):
        self.on_data_on_readers = callable
        if callable is None:
            self._set_data_available(self._ref, None)
            del self._set_functors["on_data_on_readers"]
        else:
            self._set_functors["on_data_on_readers"] = self.on_data_on_readers

            def call(subscriber, arg):
                self.on_data_on_readers(Entity.get_entity(subscriber))

            self._on_data_on_readers = _data_on_readers_fn(call)
            self._set_on_data_on_readers(self._ref, self._on_data_on_readers)

    def on_sample_lost(
        self, writer: "cyclonedds.pub.DataWriter", status: dds_c_t.sample_lost_status
    ) -> None:
        pass

    def set_on_sample_lost(
        self,
        callable: Callable[
            ["cyclonedds.pub.DataWriter", dds_c_t.sample_lost_status], None
        ],
    ):
        self.on_sample_lost = callable
        if callable is None:
            self._set_on_sample_lost(self._ref, None)
            del self._set_functors["on_sample_lost"]
        else:
            self._set_functors["on_sample_lost"] = self.on_sample_lost

            def call(writer, status, arg):
                self.on_sample_lost(Entity.get_entity(writer), status)

            self._on_sample_lost = _on_sample_lost_fn(call)
            self._set_on_sample_lost(self._ref, self._on_sample_lost)

    def on_sample_rejected(
        self,
        reader: "cyclonedds.sub.DataReader",
        status: dds_c_t.sample_rejected_status,
    ) -> None:
        pass

    def set_on_sample_rejected(
        self,
        callable: Callable[
            ["cyclonedds.sub.DataReader", dds_c_t.sample_rejected_status], None
        ],
    ):
        self.on_sample_rejected = callable
        if callable is None:
            self._set_on_sample_rejected(self._ref, None)
            del self._set_functors["on_sample_rejected"]
        else:
            self._set_functors["on_sample_rejected"] = self.on_sample_rejected

            def call(writer, status, arg):
                self.on_sample_rejected(Entity.get_entity(writer), status)

            self._on_sample_rejected = _on_sample_rejected_fn(call)
            self._set_on_sample_rejected(self._ref, self._on_sample_rejected)

    def on_requested_deadline_missed(
        self,
        reader: "cyclonedds.sub.DataReader",
        status: dds_c_t.requested_deadline_missed_status,
    ) -> None:
        pass

    def set_on_requested_deadline_missed(
        self,
        callable: Callable[
            ["cyclonedds.sub.DataReader", dds_c_t.requested_deadline_missed_status],
            None,
        ],
    ):
        self.on_requested_deadline_missed = callable
        if callable is None:
            self._set_on_requested_deadline_missed(self._ref, None)
            del self._set_functors["on_requested_deadline_missed"]
        else:
            self._set_functors[
                "on_requested_deadline_missed"
            ] = self.on_requested_deadline_missed

            def call(reader, status, arg):
                self.on_requested_deadline_missed(Entity.get_entity(reader), status)

            self._on_requested_deadline_missed = _on_requested_deadline_missed_fn(call)
            self._set_on_requested_deadline_missed(
                self._ref, self._on_requested_deadline_missed
            )

    def on_requested_incompatible_qos(
        self,
        reader: "cyclonedds.sub.DataReader",
        status: dds_c_t.requested_incompatible_qos_status,
    ) -> None:
        pass

    def set_on_requested_incompatible_qos(
        self,
        callable: Callable[
            ["cyclonedds.sub.DataReader", dds_c_t.requested_incompatible_qos_status],
            None,
        ],
    ):
        self.on_requested_incompatible_qos = callable
        if callable is None:
            self._set_on_requested_incompatible_qos(self._ref, None)
            del self._set_functors["on_requested_incompatible_qos"]
        else:
            self._set_functors[
                "on_requested_incompatible_qos"
            ] = self.on_requested_incompatible_qos

            def call(reader, status, arg):
                self.on_requested_incompatible_qos(Entity.get_entity(reader), status)

            self._on_requested_incompatible_qos = _on_requested_incompatible_qos_fn(
                call
            )
            self._set_on_requested_incompatible_qos(
                self._ref, self._on_requested_incompatible_qos
            )

    def on_publication_matched(
        self,
        writer: "cyclonedds.pub.DataWriter",
        status: dds_c_t.publication_matched_status,
    ) -> None:
        pass

    def set_on_publication_matched(
        self,
        callable: Callable[
            ["cyclonedds.pub.DataWriter", dds_c_t.publication_matched_status], None
        ],
    ):
        self.on_publication_matched = callable
        if callable is None:
            self._set_on_publication_matched(self._ref, None)
            del self._set_functors["on_publication_matched"]
        else:
            self._set_functors["on_publication_matched"] = self.on_publication_matched

            def call(writer, status, arg):
                self.on_publication_matched(Entity.get_entity(writer), status)

            self._on_publication_matched = _on_publication_matched_fn(call)
            self._set_on_publication_matched(self._ref, self._on_publication_matched)

    def on_subscription_matched(
        self,
        reader: "cyclonedds.sub.DataReader",
        status: dds_c_t.subscription_matched_status,
    ) -> None:
        pass

    def set_on_subscription_matched(
        self,
        callable: Callable[
            ["cyclonedds.sub.DataReader", dds_c_t.subscription_matched_status], None
        ],
    ):
        self.on_subscription_matched = callable
        if callable is None:
            self._set_on_subscription_matched(self._ref, None)
            del self._set_functors["on_subscription_matched"]
        else:
            self._set_functors["on_subscription_matched"] = self.on_subscription_matched

            def call(reader, status, arg):
                self.on_subscription_matched(Entity.get_entity(reader), status)

            self._on_subscription_matched = _on_subscription_matched_fn(call)
            self._set_on_subscription_matched(self._ref, self._on_subscription_matched)

    @c_call("dds_create_listener")
    def _create_listener(self, arg: ct.c_void_p) -> dds_c_t.listener_p:
        pass

    @c_call("dds_reset_listener")
    def _reset_listener(self, listener: dds_c_t.listener_p) -> None:
        pass

    @c_call("dds_copy_listener")
    def _copy_listener(self, dst: dds_c_t.listener_p, src: dds_c_t.listener_p) -> None:
        pass

    @c_call("dds_merge_listener")
    def _merge_listener(self, dst: dds_c_t.listener_p, src: dds_c_t.listener_p) -> None:
        pass

    @c_call("dds_lset_inconsistent_topic")
    def _set_inconsistent_topic(
        self, listener: dds_c_t.listener_p, callback: _inconsistent_topic_fn
    ) -> None:
        pass

    @c_call("dds_lset_data_available")
    def _set_data_available(
        self, listener: dds_c_t.listener_p, callback: _data_available_fn
    ) -> None:
        pass

    @c_call("dds_lset_liveliness_lost")
    def _set_liveliness_lost(
        self, listener: dds_c_t.listener_p, callback: _liveliness_lost_fn
    ) -> None:
        pass

    @c_call("dds_lset_liveliness_changed")
    def _set_liveliness_changed(
        self, listener: dds_c_t.listener_p, callback: _liveliness_changed_fn
    ) -> None:
        pass

    @c_call("dds_lset_offered_deadline_missed")
    def _set_on_offered_deadline_missed(
        self, listener: dds_c_t.listener_p, callback: _offered_deadline_missed_fn
    ) -> None:
        pass

    @c_call("dds_lset_offered_incompatible_qos")
    def _set_on_offered_incompatible_qos(
        self, listener: dds_c_t.listener_p, callback: _offered_incompatible_qos_fn
    ) -> None:
        pass

    @c_call("dds_lset_data_on_readers")
    def _set_on_data_on_readers(
        self, listener: dds_c_t.listener_p, callback: _data_on_readers_fn
    ) -> None:
        pass

    @c_call("dds_lset_sample_lost")
    def _set_on_sample_lost(
        self, listener: dds_c_t.listener_p, callback: _on_sample_lost_fn
    ) -> None:
        pass

    @c_call("dds_lset_sample_rejected")
    def _set_on_sample_rejected(
        self, listener: dds_c_t.listener_p, callback: _on_sample_rejected_fn
    ) -> None:
        pass

    @c_call("dds_lset_requested_deadline_missed")
    def _set_on_requested_deadline_missed(
        self, listener: dds_c_t.listener_p, callback: _on_requested_deadline_missed_fn
    ) -> None:
        pass

    @c_call("dds_lset_requested_incompatible_qos")
    def _set_on_requested_incompatible_qos(
        self, listener: dds_c_t.listener_p, callback: _on_requested_incompatible_qos_fn
    ) -> None:
        pass

    @c_call("dds_lset_publication_matched")
    def _set_on_publication_matched(
        self, listener: dds_c_t.listener_p, callback: _on_publication_matched_fn
    ) -> None:
        pass

    @c_call("dds_lset_subscription_matched")
    def _set_on_subscription_matched(
        self, listener: dds_c_t.listener_p, callback: _on_subscription_matched_fn
    ) -> None:
        pass

    @c_call("dds_delete_listener")
    def _delete_listener(self, listener: dds_c_t.listener_p) -> None:
        pass


class SampleState:
    """SampleState constants for building condition masks. This class is static and
    there should never be a need to instantiate it. It operates on the state of
    a single sample.

    Attributes
    ----------
    Read: int
        Only consider samples that have already been read.
    NotRead: int
        Only consider unread samples.
    Any: int
        Ignore the read/unread state of samples.
    """

    Read: int = 1
    NotRead: int = 2
    Any: int = 3


class ViewState:
    """ViewState constants for building condition masks. This class is static and
    there should never be a need to instantiate it. It operates on the state of
    an instance.

    Attributes
    ----------
    New: int
        Only consider samples belonging to newly created instances.
    Old: int
        Only consider samples belonging to previously created instances.
    Any: int
        Ignore the fact whether instances are new or not.
    """

    New: int = 4
    Old: int = 8
    Any: int = 12


class InstanceState:
    """InstanceState constants for building condition masks. This class is static and
    there should never be a need to instantiate it. It operates on the state of
    an instance.

    Attributes
    ----------
    Alive: int
        Only consider samples belonging to an alive instance (it has alive writer(s))
    NotAliveDisposed: int
        Only consider samples belonging to an instance that is not alive because it was actively disposed.
    NotAliveNoWriters: int
        Only consider samples belonging to an instance that is not alive because it has no writers.
    Any: int
        Ignore the liveliness status of the instance.
    """

    Alive: int = 16
    NotAliveDisposed: int = 32
    NotAliveNoWriters: int = 64
    Any: int = 112


class DDSStatus:
    """DDSStatus contains constants to build status masks. It is static and should never
    need to be instantiated.

    Attributes
    ----------
    InconsistentTopic: int
    OfferedDeadlineMissed: int
    RequestedDeadlineMissed: int
    OfferedIncompatibleQos: int
    RequestedIncompatibleQos: int
    SampleLost: int
    SampleRejected: int
    DataOnReaders: int
    DataAvailable: int
    LivelinessLost: int
    LivelinessChanged: int
    PublicationMatched: int
    SubscriptionMatched: int
    All = (1 << 14) - 1
    """

    InconsistentTopic = 1 << 0
    OfferedDeadlineMissed = 1 << 1
    RequestedDeadlineMissed = 1 << 2
    OfferedIncompatibleQos = 1 << 3
    RequestedIncompatibleQos = 1 << 4
    SampleLost = 1 << 5
    SampleRejected = 1 << 6
    DataOnReaders = 1 << 7
    DataAvailable = 1 << 8
    LivelinessLost = 1 << 9
    LivelinessChanged = 1 << 10
    PublicationMatched = 1 << 11
    SubscriptionMatched = 1 << 12
    All: int = (1 << 13) - 1


class _Condition(Entity):
    """Utility class to implement common methods between Read and Queryconditions"""

    def get_mask(self) -> int:
        mask: ct.c_uint32 = ct.c_uint32()
        ret = self._get_mask(self._ref, ct.byref(mask))
        if ret == 0:
            return mask.value
        raise DDSException(ret, f"Occurred when obtaining the mask of {repr(self)}")

    def is_triggered(self) -> bool:
        ret = self._triggered(self._ref)
        if ret < 0:
            raise DDSException(
                ret, f"Occurred when checking if {repr(self)} was triggered"
            )
        return ret == 1

    triggered: bool = property(is_triggered)

    @c_call("dds_get_mask")
    def _get_mask(
        self, condition: dds_c_t.entity, mask: ct.POINTER(ct.c_uint32)
    ) -> dds_c_t.returnv:
        pass

    @c_call("dds_triggered")
    def _triggered(self, condition: dds_c_t.entity) -> dds_c_t.returnv:
        pass


class ReadCondition(_Condition):
    """Condition that triggers when new data is available to read according to the mask.
    Construct a mask using InstanceState, ViewState and SampleState.
    """

    def __init__(self, reader: "cyclonedds.sub.DataReader", mask: int) -> None:
        """Construct a ReadCondition."""
        self.reader = reader
        self.mask = mask
        super().__init__(self._create_readcondition(reader._ref, mask))

    @c_call("dds_create_readcondition")
    def _create_readcondition(
        self, reader: dds_c_t.entity, mask: ct.c_uint32
    ) -> dds_c_t.entity:
        pass


_querycondition_filter_fn = c_callable(ct.c_bool, [ct.c_void_p])


class QueryCondition(_Condition):
    """Condition that triggers when new data is available to read according to the mask.
    Construct a mask using InstanceState, ViewState and SampleState. Add a filter function
    that receives the sample and returns a boolean whether to accept or reject the sample.
    """

    def __init__(
        self,
        reader: "cyclonedds.sub.DataReader",
        mask: int,
        filter: Callable[[Any], bool],
    ) -> None:
        """Construct a QueryCondition."""
        self.reader = reader
        self.mask = mask
        self.filter = filter

        def call(sample_pt):
            try:
                sample_info = ct.cast(sample_pt, ct.POINTER(dds_c_t.sample_buffer))[0]
                array_type = ct.c_ubyte * sample_info.len
                array = ct.cast(sample_info.buf, ct.POINTER(array_type))
                contents = array.contents[:]
                data = self.reader._topic.data_type.deserialize(bytes(contents))
                return self.filter(data)
            except Exception:  # Block any python exception from going into C
                return False

        self._filter = _querycondition_filter_fn(call)
        super().__init__(self._create_querycondition(reader._ref, mask, self._filter))

    @c_call("dds_create_querycondition")
    def _create_querycondition(
        self,
        reader: dds_c_t.entity,
        mask: ct.c_uint32,
        filter: _querycondition_filter_fn,
    ) -> dds_c_t.entity:
        pass


class GuardCondition(Entity):
    """A GuardCondition is a manually triggered condition that can be added to a :class:`WaitSet<cyclonedds.core.WaitSet>`."""

    def __init__(self, domain_participant: "cyclonedds.domain.DomainParticipant"):
        """Initialize a GuardCondition

        Parameters
        ----------
        domain_participant: DomainParticipant
            The domain in which the GuardCondition should be active.
        """
        super().__init__(self._create_guardcondition(domain_participant._ref))

    def set(self, triggered: bool) -> None:
        """Set the status of the GuardCondition to triggered or untriggered.

        Parameters
        ----------
        triggered: bool
            Wether to trigger this condition.

        Returns
        -------
        None

        Raises
        ------
        DDSException
        """
        ret = self._set_guardcondition(self._ref, triggered)
        if ret < 0:
            raise DDSException(ret, f"Occurred when calling set on {repr(self)}")

    def read(self) -> bool:
        """Read the status of the GuardCondition.

        Returns
        ----------
        bool
            Wether this condition is triggered.

        Raises
        ------
        DDSException
        """
        triggered = ct.c_bool()
        ret = self._read_guardcondition(self._ref, ct.byref(triggered))
        if ret < 0:
            raise DDSException(ret, f"Occurred when calling read on {repr(self)}")
        return bool(triggered)

    def take(self) -> bool:
        """Take the status of the GuardCondition. If it is True it will be False on the next call.

        Returns
        ----------
        bool
            Whether this condition is triggered.

        Raises
        ------
        DDSException
        """
        triggered = ct.c_bool()
        ret = self._take_guardcondition(self._ref, ct.byref(triggered))
        if ret < 0:
            raise DDSException(ret, f"Occurred when calling read on {repr(self)}")
        return bool(triggered)

    @c_call("dds_create_guardcondition")
    def _create_guardcondition(self, participant: dds_c_t.entity) -> dds_c_t.entity:
        pass

    @c_call("dds_set_guardcondition")
    def _set_guardcondition(
        self, guardcond: dds_c_t.entity, triggered: ct.c_bool
    ) -> dds_c_t.returnv:
        pass

    @c_call("dds_read_guardcondition")
    def _read_guardcondition(
        self, guardcond: dds_c_t.entity, triggered: ct.POINTER(ct.c_bool)
    ) -> dds_c_t.returnv:
        pass

    @c_call("dds_take_guardcondition")
    def _take_guardcondition(
        self, guardcond: dds_c_t.entity, triggered: ct.POINTER(ct.c_bool)
    ) -> dds_c_t.returnv:
        pass


class WaitSet(Entity):
    """A WaitSet is a way to provide synchronous access to events happening in the DDS system. You can attach almost any kind
    of entity to a WaitSet and then perform a blocking wait on the waitset. When one or more of the entities in the waitset
    trigger the wait is unblocked. What a 'trigger' is depends on the type of entity, you can find out more in
    ``todo(DDS) triggers``.
    """

    def __init__(
        self, domain_participant: "cyclonedds.domain.DomainParticipant"
    ) -> None:
        """Make a new WaitSet. It starts of empty. An empty waitset will never trigger.

        Parameters
        ----------
        domain_participant: DomainParticipant
            The domain in which you want to make a WaitSet
        """
        super().__init__(self._create_waitset(domain_participant._ref))
        self.attached = []

    def __del__(self) -> None:
        if hasattr(self, "attached"):
            for v in self.attached:
                self._waitset_detach(self._ref, v[0]._ref)
        super().__del__()

    def attach(self, entity: Entity) -> None:
        """Attach an entity to this WaitSet. This is a no-op if the entity was already attached.

        Parameters
        ----------
        entity: Entity
            The entity you wish to attach.

        Raises
        ------
        DDSException: When you try to attach a non-triggerable entity.
        """
        if self.is_attached(entity):
            return

        value_pt = ct.c_int()

        ret = self._waitset_attach(self._ref, entity._ref, ct.byref(value_pt))
        if ret < 0:
            raise DDSException(
                ret, f"Occurred when trying to attach {repr(entity)} to {repr(self)}"
            )
        self.attached.append((entity, value_pt))

    def detach(self, entity: Entity) -> None:
        """Detach an entity from this WaitSet. If it was not attach this is a no-op.
        Note that this operation is not atomic, a trigger for the detached entity could still occurr right
        after detaching it.

        Parameters
        ----------
        entity: Entity
            The entity you wish to attach

        """
        for i, v in enumerate(self.attached):
            if v[0] == entity:
                ret = self._waitset_detach(self._ref, entity._ref)
                if ret < 0:
                    raise DDSException(
                        ret,
                        f"Occurred when trying to attach {repr(entity)} to {repr(self)}",
                    )
                del self.attached[i]
                break

    def is_attached(self, entity: Entity) -> bool:
        """Check whether an entity is attached.

        Parameters
        ----------
        entity: Entity
            Check the attachment of this entity.
        """
        for v in self.attached:
            if v[0] == entity:
                return True
        return False

    def get_entities(self) -> List[Entity]:
        """Get all entities attached"""
        # Note: should spend some time on synchronisation. What if the waitset is used across threads?
        # That is probably a bad idea in python, but who is going to stop the user from doing it anyway...
        return [v[0] for v in self.attached]

    def wait(self, timeout: int) -> int:
        """Block execution and wait for one of the entities in this waitset to trigger.

        Parameters
        ----------
        timeout: int
            The maximum number of nanoseconds to block. Use the function :func:`duration<cyclonedds.util.duration>`
            to write that in a human readable format.

        Returns
        -------
        int
            The number of triggered entities. This will be 0 when a timeout occurred.
        """
        ret = self._waitset_wait(self._ref, None, 0, timeout)

        if ret >= 0:
            return ret

        raise DDSException(ret, f"Occurred while waiting in {repr(self)}")

    def wait_until(self, abstime: int) -> int:
        """Block execution and wait for one of the entities in this waitset to trigger.

        Parameters
        ----------
        abstime: int
            The absolute time in nanoseconds since the start of the program (TODO CONFIRM THIS)
            to block. Use the function :func:`duration<cyclonedds.util.duration>` to write that in
            a human readable format.

        Returns
        -------
        int
            The number of triggered entities. This will be 0 when a timeout occurred.
        """
        cs = (ct.c_void_p * len(self.attached))()
        pcs = ct.cast(cs, ct.c_void_p)
        ret = self._waitset_wait_until(
            self._ref, ct.byref(pcs), len(self.attached), abstime
        )

        if ret >= 0:
            return ret

        raise DDSException(ret, f"Occurred while waiting in {repr(self)}")

    def set_trigger(self, value: bool) -> None:
        """Manually trigger a WaitSet. It is unlikely you would need this.

        Parameters
        ----------
        value: bool
            The trigger value.
        """
        ret = self._waitset_set_trigger(self._ref, value)
        if ret < 0:
            raise DDSException(ret, f"Occurred when setting trigger in {repr(self)}")

    async def wait_async(self, timeout: Optional[int] = None) -> int:
        """Asynchronously wait for a WaitSet to trigger. Use in event-loop based applications.

        Parameters
        ----------
        timeout: int, Optional = None
            Maximum number of nanoseconds to wait before returning. By default this is infinity.
        """
        timeout = timeout or dds_infinity

        loop = asyncio.get_running_loop()
        with concurrent.futures.ThreadPoolExecutor() as pool:
            return await loop.run_in_executor(pool, self.wait, timeout)


    @c_call("dds_create_waitset")
    def _create_waitset(self, domain_participant: dds_c_t.entity) -> dds_c_t.entity:
        pass

    @c_call("dds_waitset_attach")
    def _waitset_attach(
        self, waitset: dds_c_t.entity, entity: dds_c_t.entity, x: dds_c_t.attach
    ) -> dds_c_t.returnv:
        pass

    @c_call("dds_waitset_detach")
    def _waitset_detach(
        self, waitset: dds_c_t.entity, entity: dds_c_t.entity
    ) -> dds_c_t.returnv:
        pass

    @c_call("dds_waitset_wait")
    def _waitset_wait(
        self,
        waitset: dds_c_t.entity,
        xs: ct.POINTER(dds_c_t.attach),
        nxs: ct.c_size_t,
        reltimeout: dds_c_t.duration,
    ) -> dds_c_t.returnv:
        pass

    @c_call("dds_waitset_wait_until")
    def _waitset_wait_until(
        self,
        waitset: dds_c_t.entity,
        xs: ct.POINTER(dds_c_t.attach),
        nxs: ct.c_size_t,
        abstimeout: dds_c_t.duration,
    ) -> dds_c_t.returnv:
        pass

    @c_call("dds_waitset_set_trigger")
    def _waitset_set_trigger(
        self, waitset: dds_c_t.entity, value: ct.c_bool
    ) -> dds_c_t.returnv:
        pass


class Statistics(DDS):
    """Statistics object for entity.

    Attributes
    ----------
    entity: Entity
        The handle of entity to which this set of statistics applies.
    opaque: int
        The internal data.
    time: datatime
        Time stamp of lastest call to `Statistics(entity).refresh()` in nanoseconds since epoch.
    count: int
        Number of key-value pairs.
    data: dict
        Data.

    Examples
    --------
    >>> Statistics(datawriter)
    >>> Statistics(datawriter).refresh()
    """

    entity: Entity
    opaque: int
    time: datetime
    count: int
    data: Dict[str, int]

    def __init__(self, entity: Entity):
        self.entity = entity
        self._c_statistics = self._create_statistics(entity._ref)
        if not self._c_statistics:
            raise DDSException(DDSException.DDS_RETCODE_ERROR, msg="Could not initialize statistics.")
        self._c_statistics = ct.cast(
            self._c_statistics, ct.POINTER(dds_c_t.stat_factory(self._c_statistics[0].count))
        )
        self._update()

    def __del__(self):
        self._delete_statistics(ct.cast(self._c_statistics, ct.POINTER(dds_c_t.statistics)))

    def _update(self):
        self.data = {}
        self.opaque = self._c_statistics[0].opaque
        self.time = self._c_statistics[0].time
        self.count = self._c_statistics[0].count
        self.kv = self._c_statistics[0].kv

        for i in range(self.count):
            name = self.kv[i].name.decode('utf8')  # ct.c_char_p
            value = None
            if self.kv[i].kind == stat_kind.DDS_STAT_KIND_UINT32:
                value = self.kv[i].u.u32
            elif self.kv[i].kind == stat_kind.DDS_STAT_KIND_UINT64:
                value = self.kv[i].u.u64
            elif self.kv[i].kind == stat_kind.DDS_STAT_KIND_LENGTHTIME:
                value = self.kv[i].u.lengthtime
            self.data[name] = value

    def refresh(self):
        """Update a previously created statistics structure with current values.

        Only the time stamp and the values (and "opaque") may change.
        The set of keys and the types of the values do not change.
        """
        self._refresh_statistics(ct.cast(self._c_statistics, ct.POINTER(dds_c_t.statistics)))
        self._c_statistics = ct.cast(self._c_statistics, ct.POINTER(dds_c_t.stat_factory(self._c_statistics[0].count)))
        self._update()

    def __str__(self):
        return f"Statistics({self.entity}, opaque={self.opaque}, time={self.time}, data={self.data})"

    @c_call("dds_create_statistics")
    def _create_statistics(self, entity: dds_c_t.entity) -> ct.POINTER(dds_c_t.statistics):
        pass

    @c_call("dds_refresh_statistics")
    def _refresh_statistics(self, stat: ct.POINTER(dds_c_t.statistics)) -> dds_c_t.returnv:
        pass

    @c_call("dds_delete_statistics")
    def _delete_statistics(self, stat: ct.POINTER(dds_c_t.statistics)) -> None:
        pass

    __repr__ = __str__


__all__ = [
    "DDSException",
    "Entity",
    "Qos",
    "Policy",
    "Listener",
    "DDSStatus",
    "ViewState",
    "InstanceState",
    "SampleState",
    "ReadCondition",
    "QueryCondition",
    "GuardCondition",
    "WaitSet",
    "Statistics"
]



"""
 * Copyright(c) 2021 to 2022 ZettaScale Technology and others
 *
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License v. 2.0 which is available at
 * http://www.eclipse.org/legal/epl-2.0, or the Eclipse Distribution License
 * v. 1.0 which is available at
 * http://www.eclipse.org/org/documents/edl-v10.php.
 *
 * SPDX-License-Identifier: EPL-2.0 OR BSD-3-Clause
"""

class EntityQosMapper():
    topic = {
        "Policy.Deadline",
        "Policy.DestinationOrder",
        "Policy.Durability",
        "Policy.DurabilityService",
        "Policy.History",
        "Policy.IgnoreLocal",
        "Policy.LatencyBudget",
        "Policy.Lifespan",
        "Policy.Liveliness",
        "Policy.Ownership",
        "Policy.Reliability",
        "Policy.ResourceLimits",
        "Policy.Topicdata",
        "Policy.TransportPriority"
    }

    pubsub = {
        "Policy.Groupdata",
        "Policy.IgnoreLocal",
        "Policy.Partition",
        "Policy.PresentationAccessScope"
    }

    writer = {
        "Policy.Deadline",
        "Policy.DestinationOrder",
        "Policy.Durability",
        "Policy.DurabilityService",
        "Policy.History",
        "Policy.IgnoreLocal",
        "Policy.LatencyBudget",
        "Policy.Lifespan",
        "Policy.Liveliness",
        "Policy.Ownership",
        "Policy.OwnershipStrength",
        "Policy.Reliability",
        "Policy.ResourceLimits",
        "Policy.TransportPriority",
        "Policy.Userdata",
        "Policy.WriterDataLifecycle"
    }

    reader = {
        "Policy.Deadline",
        "Policy.DestinationOrder",
        "Policy.Durability",
        "Policy.History",
        "Policy.IgnoreLocal",
        "Policy.LatencyBudget",
        "Policy.Liveliness",
        "Policy.Ownership",
        "Policy.ReaderDataLifecycle",
        "Policy.Reliability",
        "Policy.ResourceLimits",
        "Policy.TimeBasedFilter",
        "Policy.Userdata"
    }



import rich_click as click


click.rich_click.USE_RICH_MARKUP = True
click.rich_click.SHOW_ARGUMENTS = True
click.rich_click.GROUP_ARGUMENTS_OPTIONS = True
click.rich_click.COMMAND_GROUPS = {
    "cyclonedds": [
        {
            "name": "DDS entities",
            "commands": ["ls", "watch"],
        },
        {
            "name": "DDS applications",
            "commands": ["ps"],
        },
        {
            "name": "Data",
            "commands": ["subscribe", "publish"],
        },
        {
            "name": "Utilities",
            "commands": ["typeof", "performance"],
        },
    ]
}

CONTEXT_SETTINGS = dict(help_option_names=["-h", "--help"])



import sys
import pytest

from support_modules.test_tools.fixtures import FuzzingConfig

from support_modules.fuzz_tools.rand_idl.context_containers import FullContext
from support_modules.fuzz_tools.rand_idl.creator import generate_random_types
from support_modules.fuzz_tools.utility.stream import Stream, FileStream

from support_modules.fuzz_tools.checks.typeobject import check_type_object_equivalence
from support_modules.fuzz_tools.checks.keys import check_py_c_key_equivalence
from support_modules.fuzz_tools.checks.mutated import check_mutation_assignability, check_mutation_key, check_enforced_non_communication
from support_modules.fuzz_tools.checks.typebuilder import check_sertype_from_typeobj


@pytest.mark.fuzzing
def test_fuzzing_types(fuzzing_config: FuzzingConfig):
    module_name = "fuzzytypes"
    scope = generate_random_types(module_name, xcdr_version=fuzzing_config.xcdr_version, number=fuzzing_config.num_types, seed=fuzzing_config.type_seed)
    ctx = FullContext(scope)

    log = FileStream(sys.stdout)
    all_succeeded = True

    for i, typename in enumerate(ctx.topic_type_names):
        if i < fuzzing_config.skip_types:
            log << f"Testing {typename}(skipped):" << log.endl
            continue

        #log << log.endl << "[IDL]:" << log.indent << log.endl
        #log << ctx.narrow_context_of(typename).idl_file << log.endl
        #log << log.dedent
        #continue

        typelog = Stream()
        success = True
        mut_success = True
        success &= check_type_object_equivalence(typelog, ctx, typename)
        success &= check_sertype_from_typeobj(typelog, ctx, typename)

        if fuzzing_config.verbose:
            log << log.endl << "[IDL]:" << log.indent << log.endl
            log << ctx.narrow_context_of(typename).idl_file << log.endl
            log << log.dedent

        if success:
            # If python and pyc are not agreeing on keys then python and C is not so relevant.
            success = check_py_c_key_equivalence(typelog, ctx, typename, fuzzing_config.num_samples, fuzzing_config.xcdr_version)

        if success:
            # If keys are not equal we won't bother with mutations.
            mut_success &= check_mutation_assignability(typelog, ctx, typename, fuzzing_config.num_samples, fuzzing_config.xcdr_version)

        if success and mut_success:
            # If python doesn't agree with itself then C for sure won't
            mut_success &= check_mutation_key(typelog, ctx, typename, fuzzing_config.num_samples, fuzzing_config.xcdr_version)

        if success and mut_success:
            mut_success &= check_enforced_non_communication(typelog, ctx, typename, fuzzing_config.xcdr_version, fuzzing_config.verbose)

        log << f"Testing {typename}(index={i}, success={success}, mutation_success={mut_success}):" << log.endl << log.indent << typelog

        if not success or not mut_success or fuzzing_config.verbose:
            narrow_ctx = ctx.narrow_context_of(typename)
            log << log.endl << "[IDL]:" << log.indent << log.endl
            log << narrow_ctx.idl_file << log.endl
            log << log.dedent
            if fuzzing_config.store_reproducers:
                zipf, zipb = narrow_ctx.reproducer(typename, fuzzing_config.xcdr_version)
                zipf.writestr('reproducer/log.txt', typelog.string)
                zipf.close()
                with open(f"{typename}_reproducer.zip", "wb") as f:
                    f.write(zipb.getvalue())

        if fuzzing_config.mutation_failure_fatal:
            success &= mut_success

        all_succeeded &= success

        log << log.dedent

    assert all_succeeded



from dataclasses import dataclass
import pytest

import cyclonedds.internal
from cyclonedds.dynamic import get_types_for_typeid
from cyclonedds.domain import DomainParticipant
from cyclonedds.topic import Topic
from cyclonedds.pub import DataWriter
from cyclonedds.sub import DataReader
from cyclonedds.util import duration
from cyclonedds.idl import IdlStruct, types as pt

from support_modules.test_fullxcdr2_classes import XBitmask, XEnum, XStruct, XUnion


if not cyclonedds.internal.feature_type_discovery:
    pytest.skip("Skipping tests that have to do with dynamic typing since type discovery is disabled.", allow_module_level=True)


def test_dynamic_subscribe(common_setup):
    type_id = common_setup.tp.data_type.__idl__.get_type_id()

    dp = DomainParticipant(common_setup.dp.domain_id)
    datatype, _ = get_types_for_typeid(dp, type_id, duration(seconds=1))
    assert datatype

    tp = Topic(dp, common_setup.tp.name, datatype)
    dr = DataReader(dp, tp)

    common_setup.dw.write(common_setup.msg)

    assert dr.read()[0].message == common_setup.msg.message


def test_dynamic_subscribe_complex():
    dp = DomainParticipant()
    tp = Topic(dp, 'DynTest', XStruct)
    wr = DataWriter(dp, tp)

    type_id = XStruct.__idl__.get_type_id()
    datatype, tmap = get_types_for_typeid(dp, type_id, duration(seconds=1))
    assert datatype
    assert datatype.__idl__.get_type_id() == XStruct.__idl__.get_type_id()

    tp = Topic(dp, 'DynTest', datatype)
    dr = DataReader(dp, tp)

    wr.write(XStruct(A=XUnion(A=XEnum.V1), k=1))

    assert dr.read()[0].k == 1



def test_dynamic_publish_complex():
    dp = DomainParticipant()
    tp = Topic(dp, 'DynTest', XStruct)
    rd = DataReader(dp, tp)

    type_id = XStruct.__idl__.get_type_id()
    datatype, tmap = get_types_for_typeid(dp, type_id, duration(seconds=1))
    assert datatype
    assert datatype.__idl__.get_type_id() == XStruct.__idl__.get_type_id()

    tp = Topic(dp, 'DynTest', datatype)
    wr = DataWriter(dp, tp)

    wr.write(datatype(A=tmap['XUnion'](A=tmap['XEnum'].V1), k=1))

    assert rd.read()[0].k == 1


def test_dynamic_repeated_typedef_array():
    arrtype = pt.typedef["arrtype", pt.array[int, 2]]

    @dataclass
    class Foo(IdlStruct):
        foo: arrtype
        bar: arrtype

    dp = DomainParticipant()
    tp = Topic(dp, "DynTest", Foo)
    rd = DataReader(dp, tp)

    d, w = get_types_for_typeid(dp, Foo.__idl__.get_type_id(), duration(seconds=1))
    tp2 = Topic(dp, "DynTest", d)
    wr = DataWriter(dp, tp2)

    wr.write(d([1,2], [1,2]))
    assert rd.read()[0].foo[0] == 1




from dataclasses import dataclass
import time
import re
from typing import List, Optional
from cyclonedds import core, domain, builtin, dynamic, util, internal
from datetime import datetime, timedelta
from collections import defaultdict

from ..utils import LiveData
from ..idl import IdlType
from .ls_discoverables import DParticipant, DTopic, DPubSub
from .ps_discoverables import PApplication, PParticipant, PSystem
from .type_discoverables import DiscoveredType, TypeDiscoveryData


def ls_discovery(
    live: LiveData, domain_id: Optional[int], runtime: timedelta, topic: str, show_qos: bool
) -> List[DParticipant]:
    try:
        topic_re = re.compile(f"^{topic}$")
    except re.error:
        topic_re = re.compile(f"^{re.escape(topic)}$")

    if domain_id is None:
        dp = domain.DomainParticipant()
    else:
        dp = domain.DomainParticipant(domain_id)

    rdp = builtin.BuiltinDataReader(dp, builtin.BuiltinTopicDcpsParticipant)
    rcp = core.ReadCondition(
        rdp, core.SampleState.NotRead | core.ViewState.Any | core.InstanceState.Alive
    )
    rdw = builtin.BuiltinDataReader(dp, builtin.BuiltinTopicDcpsPublication)
    rcw = core.ReadCondition(
        rdw, core.SampleState.NotRead | core.ViewState.Any | core.InstanceState.Alive
    )
    rdr = builtin.BuiltinDataReader(dp, builtin.BuiltinTopicDcpsSubscription)
    rcr = core.ReadCondition(
        rdr, core.SampleState.NotRead | core.ViewState.Any | core.InstanceState.Alive
    )

    if internal.feature_topic_discovery:
        rdt = builtin.BuiltinDataReader(dp, builtin.BuiltinTopicDcpsTopic)
        rct = core.ReadCondition(
            rdt,
            core.SampleState.NotRead | core.ViewState.Any | core.InstanceState.Alive,
        )

    participants = {}
    topic_qos = {}
    start = datetime.now()
    end = start + runtime
    while datetime.now() < end and not live.terminate:
        for p in rdp.take(N=20, condition=rcp):
            if p.key in participants:
                participants[p.key].sample = p
            else:
                participants[p.key] = DParticipant(
                    sample=p, topics=[], show_qos=show_qos
                )
            if p.key == dp.guid:
                participants[p.key].is_self = True
            else:
                live.entities += 1

        if internal.feature_topic_discovery:
            for t in rdt.take(N=20, condition=rct):
                if not topic_re.match(t.topic_name):
                    continue

                for participant in participants.values():
                    for topic in participant.topics:
                        if topic.name == t.topic_name:
                            topic.qos = t.qos
                            break

                topic_qos[t.topic_name] = t.qos

        for pub in rdw.take(N=20, condition=rcw):
            if pub.participant_key != dp.guid:
                live.entities += 1

            if not topic_re.match(pub.topic_name):
                continue

            pqos = pub.qos
            naming = None
            if core.Policy.EntityName in pqos:
                naming = pqos[core.Policy.EntityName]
                pqos = pqos - core.Qos(naming)

            if pub.participant_key in participants:
                par = participants[pub.participant_key]
            else:
                par = participants[pub.participant_key] = DParticipant(
                    sample=None, topics=[], show_qos=show_qos
                )

            pub = DPubSub(endpoint=pub, qos=pqos, name=naming.name if naming else None)

            for topic in par.topics:
                if topic.name == pub.endpoint.topic_name:
                    topic.publications.append(pub)
                    break
            else:
                topic = DTopic(
                    name=pub.endpoint.topic_name,
                    subscriptions=[],
                    publications=[pub],
                    show_qos=show_qos,
                    qos=topic_qos.get(pub.endpoint.topic_name, core.Qos()),
                )
                par.topics.append(topic)

        for sub in rdr.take(N=20, condition=rcr):
            if sub.participant_key != dp.guid:
                live.entities += 1

            if not topic_re.match(sub.topic_name):
                continue

            sqos = sub.qos
            naming = None
            if core.Policy.EntityName in sqos:
                naming = sqos[core.Policy.EntityName]
                sqos = sqos - core.Qos(naming)

            if sub.participant_key in participants:
                par = participants[sub.participant_key]
            else:
                par = participants[sub.participant_key] = DParticipant(
                    sample=None, topics=[], show_qos=show_qos
                )

            sub = DPubSub(endpoint=sub, qos=sqos, name=naming.name if naming else None)

            for topic in par.topics:
                if topic.name == sub.endpoint.topic_name:
                    topic.subscriptions.append(sub)
                    break
            else:
                topic = DTopic(
                    name=sub.endpoint.topic_name,
                    subscriptions=[sub],
                    publications=[],
                    show_qos=show_qos,
                    qos=topic_qos.get(sub.endpoint.topic_name, core.Qos()),
                )
                par.topics.append(topic)

        # yield thread
        time.sleep(0.01)

    live.result = list(participants.values())
    live.delivered = True


def ps_discovery(
    live: LiveData, domain_id: Optional[int], runtime: timedelta, show_self: bool, topic: str
) -> List[PApplication]:
    try:
        topic_re = re.compile(f"^{topic}$")
    except re.error:
        topic_re = re.compile(f"^{re.escape(topic)}$")

    if domain_id is None:
        dp = domain.DomainParticipant()
    else:
        dp = domain.DomainParticipant(domain_id)

    rdp = builtin.BuiltinDataReader(dp, builtin.BuiltinTopicDcpsParticipant)
    rcp = core.ReadCondition(
        rdp, core.SampleState.NotRead | core.ViewState.Any | core.InstanceState.Alive
    )
    rdw = builtin.BuiltinDataReader(dp, builtin.BuiltinTopicDcpsPublication)
    rcw = core.ReadCondition(
        rdw, core.SampleState.NotRead | core.ViewState.Any | core.InstanceState.Alive
    )
    rdr = builtin.BuiltinDataReader(dp, builtin.BuiltinTopicDcpsSubscription)
    rcr = core.ReadCondition(
        rdr, core.SampleState.NotRead | core.ViewState.Any | core.InstanceState.Alive
    )

    applications = {}
    participants = {}

    hostname_get = core.Policy.Property("__Hostname", "")
    appname_get = core.Policy.Property("__ProcessName", "")
    pid_get = core.Policy.Property("__Pid", "")
    address_get = core.Policy.Property("__NetworkAddresses", "")

    start = datetime.now()
    end = start + runtime
    while datetime.now() < end and not live.terminate:
        for p in rdp.take(N=20, condition=rcp):
            if p.key == dp.guid and not show_self:
                continue

            hostname = (
                p.qos[hostname_get].value
                if p.qos[hostname_get] is not None
                else "Unknown"
            )
            appname = (
                p.qos[appname_get].value
                if p.qos[appname_get] is not None
                else "Unknown"
            )
            pid = p.qos[pid_get].value if p.qos[pid_get] is not None else "Unknown"
            address = (
                p.qos[address_get].value
                if p.qos[address_get] is not None
                else "Unknown"
            )

            key = f"{hostname}.{appname}.{pid}"
            name = (
                p.qos[core.Policy.EntityName].name
                if core.Policy.EntityName in p.qos
                else None
            )
            participant = PParticipant(name=name, key=p.key)
            live.entities += 1

            if p.key in participants:
                participant.topics = participants[p.key].topics

            participants[p.key] = participant

            if key in applications:
                applications[key].participants.append(participant)
            else:
                applications[key] = PApplication(
                    hostname=hostname,
                    appname=appname,
                    pid=pid,
                    addresses=address,
                    participants=[participant],
                )

        for pub in rdw.take(N=20, condition=rcw):
            if pub.participant_key == dp.guid:
                continue
            live.entities += 1

            if not topic_re.match(pub.topic_name):
                continue

            if pub.participant_key in participants:
                par = participants[pub.participant_key]
            else:
                par = participants[pub.participant_key] = PParticipant(
                    name=None, key=pub.participant_key
                )

            par.topics.add(pub.topic_name)

        for sub in rdr.take(N=20, condition=rcr):
            if sub.participant_key == dp.guid:
                continue
            live.entities += 1

            if not topic_re.match(sub.topic_name):
                continue

            if sub.participant_key in participants:
                par = participants[sub.participant_key]
            else:
                par = participants[sub.participant_key] = PParticipant(
                    name=None, key=sub.participant_key
                )

            par.topics.add(sub.topic_name)

        # yield thread
        time.sleep(0.01)

    live.result = PSystem(list(applications.values()))
    live.delivered = True


def type_discovery(
    live: LiveData, domain_id: Optional[int], runtime: timedelta, topic: str
) -> List[PApplication]:
    try:
        topic_re = re.compile(f"^{topic}$")
    except re.error:
        topic_re = re.compile(f"^{re.escape(topic)}$")

    if domain_id is None:
        dp = domain.DomainParticipant()
    else:
        dp = domain.DomainParticipant(domain_id)

    rdw = builtin.BuiltinDataReader(dp, builtin.BuiltinTopicDcpsPublication)
    rcw = core.ReadCondition(
        rdw, core.SampleState.NotRead | core.ViewState.Any | core.InstanceState.Alive
    )
    rdr = builtin.BuiltinDataReader(dp, builtin.BuiltinTopicDcpsSubscription)
    rcr = core.ReadCondition(
        rdr, core.SampleState.NotRead | core.ViewState.Any | core.InstanceState.Alive
    )

    if internal.feature_topic_discovery:
        rdt = builtin.BuiltinDataReader(dp, builtin.BuiltinTopicDcpsTopic)
        rct = core.ReadCondition(
            rdt,
            core.SampleState.NotRead | core.ViewState.Any | core.InstanceState.Alive,
        )

    discovery_data = TypeDiscoveryData()

    start = datetime.now()
    end = start + runtime
    while datetime.now() < end and not live.terminate:
        for t in rdt.take(N=20, condition=rct):
            if topic_re.match(t.topic_name):
                discovery_data.topic_qosses.append(t.qos)

        for pub in rdw.take(N=20, condition=rcw):
            if topic_re.match(pub.topic_name):
                if pub.type_id is not None:
                    discovery_data.add_type_id(str(pub.participant_key), pub.type_id)
                discovery_data.writer_qosses.append(pub.qos)
                live.entities += 1

        for sub in rdr.take(N=20, condition=rcr):
            if topic_re.match(sub.topic_name):
                if sub.type_id is not None:
                    discovery_data.add_type_id(str(sub.participant_key), sub.type_id)
                discovery_data.reader_qosses.append(sub.qos)
                live.entities += 1

        # yield thread
        time.sleep(0.01)

    for type_id, discovered_type in discovery_data.types.items():
        datatype, all_nested_datatypes = dynamic.get_types_for_typeid(
            dp, type_id, util.duration(seconds=runtime.total_seconds())
        )

        discovered_type.code = IdlType.idl([datatype])
        discovered_type.dtype = datatype
        discovered_type.nested_dtypes = all_nested_datatypes

    live.result = discovery_data
    live.delivered = True



import pytest
import random

import support_modules.test_rec_classes as trc


def test_recursive_types():
    v1 = trc.Node(
        left=trc.OptNode(node=trc.Node(
            left=trc.OptNode(nothing=None),
            right=trc.OptNode(node=trc.Node(
                left=trc.OptNode(nothing=None),
                right=trc.OptNode(nothing=None),
                value=12
            )),
            value = 6
        )),
        right=trc.OptNode(node=trc.Node(
            left=trc.OptNode(nothing=None),
            right=trc.OptNode(node=trc.Node(
                left=trc.OptNode(nothing=None),
                right=trc.OptNode(nothing=None),
                value=-12
            )),
            value = -6
        )),
        value=0
    )
    v1 = trc.Node.deserialize(v1.serialize())
    assert v1 == trc.Node.deserialize(v1.serialize())


def test_optional_recursive():
    tree = trc.CNode(value=0)

    l = list(range(-100, 100))
    random.seed(0)
    random.shuffle(l)
    for i in l:
        tree.add(i)

    assert tree == trc.CNode.deserialize(tree.serialize())




import time
from typing import Any, Optional
from cyclonedds import qos, domain, sub, topic

from ..utils import LiveData


def subscribe(
    live: LiveData,
    domain_id: Optional[int],
    topic_name: str,
    datatype: Any,
    topicqos: qos.Qos,
    readerqos: qos.Qos,
):
    if domain_id is None:
        dp = domain.DomainParticipant()
    else:
        dp = domain.DomainParticipant(domain_id)
    tp = topic.Topic(dp, topic_name, datatype, qos=topicqos)
    rd = sub.DataReader(dp, tp, qos=readerqos)

    while not live.terminate:
        for data in rd.take(N=20):
            live.printables.put(data, block=True)

        # yield thread
        time.sleep(0.01)

    live.delivered = True
